=== modified file 'Makefile'
--- Makefile	2008-07-20 03:51:41 +0000
+++ Makefile	2008-07-29 03:35:39 +0000
@@ -1,7 +1,31 @@
-CFLAGS=-Wall -g -std=gnu99
-LDFLAGS=-lgnutls
-
-all: plugbasedclient
-
+WARN=-O -Wall -Wformat=2 -Winit-self -Wmissing-include-dirs -Wswitch-default -Wswitch-enum -Wunused-parameter -Wstrict-aliasing=2 -Wextra -Wfloat-equal -Wundef -Wshadow -Wunsafe-loop-optimizations -Wpointer-arith -Wbad-function-cast -Wcast-qual -Wcast-align -Wwrite-strings -Wconversion -Wstrict-prototypes -Wold-style-definition -Wpacked -Wnested-externs -Wunreachable-code -Winline -Wvolatile-register-var 
+DEBUG=-ggdb3
+# For info about _FORTIFY_SOURCE, see
+# <http://gcc.gnu.org/ml/gcc-patches/2004-09/msg02055.html>
+FORTIFY=-D_FORTIFY_SOURCE=2 # -fstack-protector-all
+#COVERAGE=--coverage
+OPTIMIZE=-Os
+LANGUAGE=-std=gnu99
+
+# Do not change these two
+CFLAGS=$(WARN) $(DEBUG) $(FORTIFY) $(COVERAGE) $(OPTIMIZE) $(LANGUAGE)
+LDFLAGS=$(COVERAGE)
+
+PROGS=plugbasedclient plugins.d/mandosclient plugins.d/passprompt
+
+objects=$(shell for p in $(PROGS); do echo $${p}.o; done)
+
+all: $(PROGS)
+
+plugbasedclient: plugbasedclient.o
+	$(LINK.o) -lgnutls $(COMMON) $^ $(LOADLIBES) $(LDLIBS) -o $@
+
+plugins.d/mandosclient: plugins.d/mandosclient.o
+	$(LINK.o) -lgnutls -lavahi-core -lgpgme $(COMMON) $^ $(LOADLIBES) $(LDLIBS) -o $@
+
+plugins.d/passprompt: plugins.d/passprompt.o
+	$(LINK.o) $(COMMON) $^ $(LOADLIBES) $(LDLIBS) -o $@
+
+.PHONY : clean
 clean:
-	rm -f plugbasedclient
+	-rm -f $(PROGS) $(objects) core

=== modified file 'TODO'
--- TODO	2008-07-20 06:33:48 +0000
+++ TODO	2008-08-03 21:45:14 +0000
@@ -1,12 +1,37 @@
-[Client]
-configuration for OpenPGP key and interface
-IPv4 support
-
-[Server]
-config file
-run-time communication with server
-
-[Mandos-tools/utilities]
-	List clients
-	Enable client
-	Disable client
+-*- org -*-
+
+* Mandos client
+** Use argp instead of getopt_long
+** header files/symbols tally
+** check exit codes of all system calls
+** [#C] IPv4 support
+** use strsep instead of strtok?
+** Use argp
+
+* Passprompt
+** Use argp
+
+* Pluginbasedclient
+** header files/symbols tally
+** check exit codes of all system calls
+** use strsep instead of strtok?
+** use config file in addition to arguments
+** [#C] pass things in environment, like device name, etc
+
+* Server
+** config for:
+*** TXT record
+** protocol version header
+** Run-time communication with server
+   probably using D-Bus
+
+* Mandos-tools/utilities
+  All of this probably using D-Bus
+** List clients
+** Enable client
+** Disable client
+
+* Installer
+
+
+#+STARTUP: showall

=== renamed file 'mandos-clients.conf' => 'clients.conf'
--- mandos-clients.conf	2008-07-20 06:33:48 +0000
+++ clients.conf	2008-07-29 03:35:39 +0000
@@ -1,3 +1,8 @@
+[DEFAULT]
+timeout = 1h
+interval = 5m
+checker = fping -q -- %%(fqdn)s
+
 # Example
 [foo]
 fingerprint = 3e393aeaefb84c7e89e2f547b3a107558fca3a27

=== modified file 'plugbasedclient.c'
--- plugbasedclient.c	2008-07-20 03:51:41 +0000
+++ plugbasedclient.c	2008-08-03 03:33:56 +0000
@@ -1,17 +1,56 @@
-#include <stdio.h>	/* popen, fileno */
-#include <iso646.h>	/* and, or, not */
-#include <sys/types.h>	/* DIR, opendir, stat, struct stat, waitpid,
-			   WIFEXITED, WEXITSTATUS, wait */
-#include <sys/wait.h>	/* wait */
-#include <dirent.h>	/* DIR, opendir */
-#include <sys/stat.h>	/* stat, struct stat */
-#include <unistd.h>	/* stat, struct stat, chdir */
-#include <stdlib.h>	/* EXIT_FAILURE */
-#include <sys/select.h>	/* fd_set, select, FD_ZERO, FD_SET, FD_ISSET */
-#include <string.h>	/* strlen, strcpy, strcat */
-#include <stdbool.h>	/* true */
-#include <sys/wait.h>	/* waitpid, WIFEXITED, WEXITSTATUS */
-#include <errno.h>	/* errno */
+/*  -*- coding: utf-8 -*- */
+/*
+ * Mandos plugin runner - Run Mandos plugins
+ *
+ * Copyright © 2007-2008 Teddy Hogeborn & Björn Påhlsson
+ * 
+ * This program is free software: you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation, either version 3 of the
+ * License, or (at your option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, see
+ * <http://www.gnu.org/licenses/>.
+ * 
+ * Contact the authors at <mandos@fukt.bsnet.se>.
+ */
+
+#define _GNU_SOURCE		/* TEMP_FAILURE_RETRY() */
+
+#include <stdio.h>		/* popen(), fileno(), fprintf(),
+				   stderr, STDOUT_FILENO */
+#include <iso646.h>		/* and, or, not */
+#include <sys/types.h>	        /* DIR, opendir(), stat(),
+				   struct stat, waitpid(),
+				   WIFEXITED(), WEXITSTATUS(),
+				   wait() */
+#include <sys/wait.h>		/* wait() */
+#include <dirent.h>		/* DIR, struct dirent, opendir(),
+				   readdir(), closedir() */
+#include <sys/stat.h>		/* struct stat, stat(), S_ISREG() */
+#include <unistd.h>		/* struct stat, stat(), S_ISREG(),
+				   fcntl() */
+#include <fcntl.h>		/* fcntl() */
+#include <stddef.h>		/* NULL */
+#include <stdlib.h>		/* EXIT_FAILURE */
+#include <sys/select.h>		/* fd_set, select(), FD_ZERO(),
+				   FD_SET(), FD_ISSET() */
+#include <string.h>		/* strlen(), strcpy(), strcat() */
+#include <stdbool.h>		/* true */
+#include <sys/wait.h>		/* waitpid(), WIFEXITED(),
+				   WEXITSTATUS() */
+#include <errno.h>		/* errno */
+#include <argp.h>		/* struct argp_option,
+				   struct argp_state, struct argp,
+				   argp_parse() */
+
+#define BUFFER_SIZE 256
 
 struct process;
 
@@ -19,31 +58,296 @@
   pid_t pid;
   int fd;
   char *buffer;
-  int buffer_size;
-  int buffer_length;
+  size_t buffer_size;
+  size_t buffer_length;
+  bool eof;
+  bool completed;
+  int status;
   struct process *next;
 } process;
 
-#define BUFFER_SIZE 256
+typedef struct plugin{
+  char *name;			/* can be NULL or any plugin name */
+  char **argv;
+  int argc;
+  bool disabled;
+  struct plugin *next;
+} plugin;
+
+static plugin *getplugin(char *name, plugin **plugin_list){
+  for (plugin *p = *plugin_list; p != NULL; p = p->next){
+    if ((p->name == name)
+	or (p->name and name and (strcmp(p->name, name) == 0))){
+      return p;
+    }
+  }
+  /* Create a new plugin */
+  plugin *new_plugin = malloc(sizeof(plugin));
+  if (new_plugin == NULL){
+    perror("malloc");
+    exit(EXIT_FAILURE);
+  }
+  new_plugin->name = name;
+  new_plugin->argv = malloc(sizeof(char *) * 2);
+  if (new_plugin->argv == NULL){
+    perror("malloc");
+    exit(EXIT_FAILURE);
+  }
+  new_plugin->argv[0] = name;
+  new_plugin->argv[1] = NULL;
+  new_plugin->argc = 1;
+  new_plugin->disabled = false;
+  new_plugin->next = *plugin_list;
+  /* Append the new plugin to the list */
+  *plugin_list = new_plugin;
+  return new_plugin;
+}
+
+static void addargument(plugin *p, char *arg){
+  p->argv[p->argc] = arg;
+  p->argv = realloc(p->argv, sizeof(char *) * (size_t)(p->argc + 2));
+  if (p->argv == NULL){
+    perror("malloc");
+    exit(EXIT_FAILURE);
+  }
+  p->argc++;
+  p->argv[p->argc] = NULL;
+}
+
+/*
+ * Based on the example in the GNU LibC manual chapter 13.13 "File
+ * Descriptor Flags".
+ * *Note File Descriptor Flags:(libc)Descriptor Flags.
+ */
+static int set_cloexec_flag(int fd)
+{
+  int ret = fcntl(fd, F_GETFD, 0);
+  /* If reading the flags failed, return error indication now. */
+  if(ret < 0){
+    return ret;
+  }
+  /* Store modified flag word in the descriptor. */
+  return fcntl(fd, F_SETFD, ret | FD_CLOEXEC);
+}
+
+const char *argp_program_version = "plugbasedclient 0.9";
+const char *argp_program_bug_address = "<mandos@fukt.bsnet.se>";
+
+process *process_list = NULL;
+
+/* Mark a process as completed when it exits, and save its exit
+   status. */
+void handle_sigchld(__attribute__((unused)) int sig){
+  process *proc = process_list;
+  int status;
+  pid_t pid = wait(&status);
+  while(proc != NULL and proc->pid != pid){
+    proc = proc->next;
+  }
+  if(proc == NULL){
+    /* Process not found in process list */
+    return;
+  }
+  proc->status = status;
+  proc->completed = true;
+}
 
 int main(int argc, char *argv[]){
-  char plugindir[] = "plugins.d";
-  size_t d_name_len, plugindir_len = sizeof(plugindir)-1;
-  DIR *dir;
+  const char *plugindir = "/conf/conf.d/mandos/plugins.d";
+  size_t d_name_len;
+  DIR *dir = NULL;
   struct dirent *dirst;
   struct stat st;
-  fd_set rfds_orig;
+  fd_set rfds_all;
   int ret, maxfd = 0;
-  process *process_list = NULL;
+  uid_t uid = 65534;
+  gid_t gid = 65534;
+  bool debug = false;
+  int exitstatus = EXIT_SUCCESS;
+  struct sigaction old_sigchld_action;
+  struct sigaction sigchld_action = { .sa_handler = handle_sigchld,
+				      .sa_flags = SA_NOCLDSTOP };
+  char *plus_options = NULL;
+  char **plus_argv = NULL;
+  
+  /* Establish a signal handler */
+  sigemptyset(&sigchld_action.sa_mask);
+  ret = sigaddset(&sigchld_action.sa_mask, SIGCHLD);
+  if(ret < 0){
+    perror("sigaddset");
+    exit(EXIT_FAILURE);
+  }
+  ret = sigaction(SIGCHLD, &sigchld_action, &old_sigchld_action);
+  if(ret < 0){
+    perror("sigaction");
+    exit(EXIT_FAILURE);
+  }
+  
+  /* The options we understand. */
+  struct argp_option options[] = {
+    { .name = "global-options", .key = 'g',
+      .arg = "OPTION[,OPTION[,...]]",
+      .doc = "Options passed to all plugins" },
+    { .name = "options-for", .key = 'o',
+      .arg = "PLUGIN:OPTION[,OPTION[,...]]",
+      .doc = "Options passed only to specified plugin" },
+    { .name = "disable", .key = 'd',
+      .arg = "PLUGIN",
+      .doc = "Disable a specific plugin", .group = 1 },
+    { .name = "plugin-dir", .key = 128,
+      .arg = "DIRECTORY",
+      .doc = "Specify a different plugin directory", .group = 2 },
+    { .name = "userid", .key = 129,
+      .arg = "ID", .flags = 0,
+      .doc = "User ID the plugins will run as", .group = 2 },
+    { .name = "groupid", .key = 130,
+      .arg = "ID", .flags = 0,
+      .doc = "Group ID the plugins will run as", .group = 2 },
+    { .name = "debug", .key = 131,
+      .doc = "Debug mode", .group = 3 },
+    { .name = NULL }
+  };
+  
+  error_t parse_opt (int key, char *arg, struct argp_state *state) {
+    /* Get the INPUT argument from `argp_parse', which we know is a
+       pointer to our plugin list pointer. */
+    plugin **plugins = state->input;
+    switch (key) {
+    case 'g':
+      if (arg != NULL){
+	char *p = strtok(arg, ",");
+	do{
+	  addargument(getplugin(NULL, plugins), p);
+	  p = strtok(NULL, ",");
+	} while (p != NULL);
+      }
+      break;
+    case 'o':
+      if (arg != NULL){
+	char *name = strtok(arg, ":");
+	char *p = strtok(NULL, ":");
+	if(p != NULL){
+	  p = strtok(p, ",");
+	  do{
+	    addargument(getplugin(name, plugins), p);
+	    p = strtok(NULL, ",");
+	  } while (p != NULL);
+	}
+      }
+      break;
+    case 'd':
+      if (arg != NULL){
+	getplugin(arg, plugins)->disabled = true;
+      }
+      break;
+    case 128:
+      plugindir = arg;
+      break;
+    case 129:
+      uid = (uid_t)strtol(arg, NULL, 10);
+      break;
+    case 130:
+      gid = (gid_t)strtol(arg, NULL, 10);
+      break;
+    case 131:
+      debug = true;
+      break;
+    case ARGP_KEY_ARG:
+      if(plus_options != NULL or arg == NULL or arg[0] != '+'){
+	argp_usage (state);
+      }
+      plus_options = arg;
+      break;
+    case ARGP_KEY_END:
+      break;
+    default:
+      return ARGP_ERR_UNKNOWN;
+    }
+    return 0;
+  }
+  
+  plugin *plugin_list = NULL;
+  
+  struct argp argp = { .options = options, .parser = parse_opt,
+		       .args_doc = "[+PLUS_SEPARATED_OPTIONS]",
+		       .doc = "Mandos plugin runner -- Run plugins" };
+  
+  argp_parse (&argp, argc, argv, 0, 0, &plugin_list);  
+  
+  if(plus_options){
+    /* This is a mangled argument in the form of
+     "+--option+--other-option=parameter+--yet-another-option", etc */
+    /* Make new argc and argv vars, and call argp_parse() again. */
+    plus_options++;		/* skip the first '+' character */
+    const char delims[] = "+";
+    char *arg;
+    int new_argc = 1;
+    plus_argv = malloc(sizeof(char*) * 2);
+    if(plus_argv == NULL){
+      perror("malloc");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    plus_argv[0] = argv[0];
+    plus_argv[1] = NULL;
+    arg = strtok(plus_options, delims); /* Get first argument */
+    while(arg != NULL){
+      new_argc++;
+      plus_argv = realloc(plus_argv, sizeof(char *)
+			 * ((unsigned int) new_argc + 1));
+      if(plus_argv == NULL){
+	perror("malloc");
+	exitstatus = EXIT_FAILURE;
+	goto end;
+      }
+      plus_argv[new_argc-1] = arg;
+      plus_argv[new_argc] = NULL;
+      arg = strtok(NULL, delims); /* Get next argument */
+    }
+    argp_parse (&argp, new_argc, plus_argv, 0, 0, &plugin_list);  
+  }
+  
+  if(debug){
+    for(plugin *p = plugin_list; p != NULL; p=p->next){
+      fprintf(stderr, "Plugin: %s has %d arguments\n",
+	      p->name ? p->name : "Global", p->argc - 1);
+      for(char **a = p->argv; *a != NULL; a++){
+	fprintf(stderr, "\tArg: %s\n", *a);
+      }
+    }
+  }
+  
+  ret = setuid(uid);
+  if (ret == -1){
+    perror("setuid");
+  }
+  
+  setgid(gid);
+  if (ret == -1){
+    perror("setgid");
+  }
   
   dir = opendir(plugindir);
-
   if(dir == NULL){
-    fprintf(stderr, "Can not open directory\n");
-    return EXIT_FAILURE;
-  }
-  
-  FD_ZERO(&rfds_orig);
+    perror("Could not open plugin dir");
+    exitstatus = EXIT_FAILURE;
+    goto end;
+  }
+  
+  /* Set the FD_CLOEXEC flag on the directory, if possible */
+  {
+    int dir_fd = dirfd(dir);
+    if(dir_fd >= 0){
+      ret = set_cloexec_flag(dir_fd);
+      if(ret < 0){
+	perror("set_cloexec_flag");
+	exitstatus = EXIT_FAILURE;
+	goto end;
+      }
+    }
+  }
+  
+  FD_ZERO(&rfds_all);
   
   while(true){
     dirst = readdir(dir);
@@ -55,128 +359,338 @@
     
     d_name_len = strlen(dirst->d_name);
     
-    // Ignore dotfiles and backup files
-    if (dirst->d_name[0] == '.'
-	or dirst->d_name[d_name_len - 1] == '~'){
-      continue;
-    }
-
-    char *filename = malloc(d_name_len + plugindir_len + 2);
-    strcpy(filename, plugindir);
-    strcat(filename, "/");
-    strcat(filename, dirst->d_name);    
-
+    // Ignore dotfiles, backup files and other junk
+    {
+      bool bad_name = false;
+      
+      const char const *bad_prefixes[] = { ".", "#", NULL };
+      
+      const char const *bad_suffixes[] = { "~", "#", ".dpkg-new",
+					   ".dpkg-old",
+					   ".dpkg-divert", NULL };
+      for(const char **pre = bad_prefixes; *pre != NULL; pre++){
+	size_t pre_len = strlen(*pre);
+	if((d_name_len >= pre_len)
+	   and strncmp((dirst->d_name), *pre, pre_len) == 0){
+	  if(debug){
+	    fprintf(stderr, "Ignoring plugin dir entry \"%s\""
+		    " with bad prefix %s\n", dirst->d_name, *pre);
+	  }
+	  bad_name = true;
+	  break;
+	}
+      }
+      
+      if(bad_name){
+	continue;
+      }
+      
+      for(const char **suf = bad_suffixes; *suf != NULL; suf++){
+	size_t suf_len = strlen(*suf);
+	if((d_name_len >= suf_len)
+	   and (strcmp((dirst->d_name)+d_name_len-suf_len, *suf)
+		== 0)){
+	  if(debug){
+	    fprintf(stderr, "Ignoring plugin dir entry \"%s\""
+		    " with bad suffix %s\n", dirst->d_name, *suf);
+	  }
+	  bad_name = true;
+	  break;
+	}
+      }
+      
+      if(bad_name){
+	continue;
+      }
+    }
+    
+    char *filename = malloc(d_name_len + strlen(plugindir) + 2);
+    if (filename == NULL){
+      perror("malloc");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    strcpy(filename, plugindir); /* Spurious warning */
+    strcat(filename, "/");	/* Spurious warning */
+    strcat(filename, dirst->d_name); /* Spurious warning */
+    
     stat(filename, &st);
-
-    if (S_ISREG(st.st_mode) and (access(filename, X_OK) == 0)){
-      // Starting a new process to be watched
-      process *new_process = malloc(sizeof(process));
-      int pipefd[2];
-      pipe(pipefd);
-      new_process->pid = fork();
-      if(new_process->pid == 0){
-	/* this is the child process */
+    
+    if (not S_ISREG(st.st_mode)	or (access(filename, X_OK) != 0)){
+      if(debug){
+	fprintf(stderr, "Ignoring plugin dir entry \"%s\""
+		" with bad type or mode\n", filename);
+      }
+      free(filename);
+      continue;
+    }
+    if(getplugin(dirst->d_name, &plugin_list)->disabled){
+      if(debug){
+	fprintf(stderr, "Ignoring disabled plugin \"%s\"\n",
+		dirst->d_name);
+      }
+      free(filename);
+      continue;
+    }
+    plugin *p = getplugin(dirst->d_name, &plugin_list);
+    {
+      /* Add global arguments to argument list for this plugin */
+      plugin *g = getplugin(NULL, &plugin_list);
+      for(char **a = g->argv + 1; *a != NULL; a++){
+	addargument(p, *a);
+      }
+    }
+    int pipefd[2]; 
+    ret = pipe(pipefd);
+    if (ret == -1){
+      perror("pipe");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    ret = set_cloexec_flag(pipefd[0]);
+    if(ret < 0){
+      perror("set_cloexec_flag");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    ret = set_cloexec_flag(pipefd[1]);
+    if(ret < 0){
+      perror("set_cloexec_flag");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    /* Block SIGCHLD until process is safely in process list */
+    ret = sigprocmask (SIG_BLOCK, &sigchld_action.sa_mask, NULL);
+    if(ret < 0){
+      perror("sigprocmask");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    // Starting a new process to be watched
+    pid_t pid = fork();
+    if(pid == 0){
+      /* this is the child process */
+      ret = sigaction(SIGCHLD, &old_sigchld_action, NULL);
+      if(ret < 0){
+	perror("sigaction");
+	_exit(EXIT_FAILURE);
+      }
+      ret = sigprocmask (SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
+      if(ret < 0){
+	perror("sigprocmask");
+	_exit(EXIT_FAILURE);
+      }
+      dup2(pipefd[1], STDOUT_FILENO); /* replace our stdout */
+      
+      if(dirfd(dir) < 0){
+	/* If dir has no file descriptor, we could not set FD_CLOEXEC
+	   above and must now close it manually here. */
 	closedir(dir);
-	close(pipefd[0]);	/* close unused read end of pipe */
-	dup2(pipefd[1], STDOUT_FILENO); /* replace our stdout */
-	/* create a new modified argument list */
-	char **new_argv = malloc(sizeof(char *) * (argc + 1));
-	new_argv[0] = filename;
-	for(int i = 1; i < argc; i++){
-	  new_argv[i] = argv[i];
-	}
-	new_argv[argc] = NULL;
-	if(execv(filename, new_argv) < 0){
-	  perror(argv[0]);
-	  close(pipefd[1]);
-	  exit(EXIT_FAILURE);
-	}
-	/* no return */
-      }
-      close(pipefd[1]);		/* close unused write end of pipe */
-      new_process->fd = pipefd[0];
-      new_process->buffer = malloc(BUFFER_SIZE);
-      if (new_process->buffer == NULL){
-	perror(argv[0]);
-	goto end;
-      }
-      new_process->buffer_size = BUFFER_SIZE;
-      new_process->buffer_length = 0;
-      FD_SET(new_process->fd, &rfds_orig);
-      
-      if (maxfd < new_process->fd){
-	maxfd = new_process->fd;
-      }
-      
-      //List handling
-      new_process->next = process_list;
-      process_list = new_process;
-    }
+      }
+      if(execv(filename, p->argv) < 0){
+	perror("execv");
+	_exit(EXIT_FAILURE);
+      }
+      /* no return */
+    }
+    /* parent process */
+    free(filename);
+    close(pipefd[1]);		/* close unused write end of pipe */
+    process *new_process = malloc(sizeof(process));
+    if (new_process == NULL){
+      perror("malloc");
+      ret = sigprocmask (SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
+      if(ret < 0){
+	perror("sigprocmask");
+      }
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    
+    *new_process = (struct process){ .pid = pid,
+				     .fd = pipefd[0],
+				     .next = process_list };
+    // List handling
+    process_list = new_process;
+    /* Unblock SIGCHLD so signal handler can be run if this process
+       has already completed */
+    ret = sigprocmask (SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
+    if(ret < 0){
+      perror("sigprocmask");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    
+    FD_SET(new_process->fd, &rfds_all);
+    
+    if (maxfd < new_process->fd){
+      maxfd = new_process->fd;
+    }
+    
+  }
+  
+  /* Free the plugin list */
+  for(plugin *next; plugin_list != NULL; plugin_list = next){
+    next = plugin_list->next;
+    free(plugin_list->argv);
+    free(plugin_list);
   }
   
   closedir(dir);
+  dir = NULL;
   
-  if (process_list != NULL){
-    while(true){
-      fd_set rfds = rfds_orig;
-      int select_ret = select(maxfd+1, &rfds, NULL, NULL, NULL);
-      if (select_ret == -1){
-	perror(argv[0]);
-	goto end;
-      }else{	
-	for(process *process_itr = process_list; process_itr != NULL;
-	    process_itr = process_itr->next){
-	  if(FD_ISSET(process_itr->fd, &rfds)){
-	    if(process_itr->buffer_length + BUFFER_SIZE
-	       > process_itr->buffer_size){
-		process_itr->buffer = realloc(process_itr->buffer,
-					      process_itr->buffer_size
-					      + BUFFER_SIZE);
-		if (process_itr->buffer == NULL){
-		  perror(argv[0]);
-		  goto end;
-		}
-		process_itr->buffer_size += BUFFER_SIZE;
+  if (process_list == NULL){
+    fprintf(stderr, "No plugin processes started, exiting\n");
+    exitstatus = EXIT_FAILURE;
+    goto end;
+  }
+  while(process_list){
+    fd_set rfds = rfds_all;
+    int select_ret = select(maxfd+1, &rfds, NULL, NULL, NULL);
+    if (select_ret == -1){
+      perror("select");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    /* OK, now either a process completed, or something can be read
+       from one of them */
+    for(process *proc = process_list; proc ; proc = proc->next){
+      /* Is this process completely done? */
+      if(proc->eof and proc->completed){
+	/* Only accept the plugin output if it exited cleanly */
+	if(not WIFEXITED(proc->status)
+	   or WEXITSTATUS(proc->status) != 0){
+	  /* Bad exit by plugin */
+	  if(debug){
+	    if(WIFEXITED(proc->status)){
+	      fprintf(stderr, "Plugin %d exited with status %d\n",
+		      proc->pid, WEXITSTATUS(proc->status));
+	    } else if(WIFSIGNALED(proc->status)) {
+	      fprintf(stderr, "Plugin %d killed by signal %d\n",
+		      proc->pid, WTERMSIG(proc->status));
+	    } else if(WCOREDUMP(proc->status)){
+	      fprintf(stderr, "Plugin %d dumped core\n", proc->pid);
 	    }
-	    ret = read(process_itr->fd, process_itr->buffer
-		       + process_itr->buffer_length, BUFFER_SIZE);
-	    process_itr->buffer_length+=ret;
-	    if(ret == 0){
-	      /* got EOF */
-	      /* wait for process exit */
-	      int status;
-	      waitpid(process_itr->pid, &status, 0);
-	      if(WIFEXITED(status) and WEXITSTATUS(status) == 0){
-		write(STDOUT_FILENO, process_itr->buffer,
-		      process_itr->buffer_length);
-		goto end;
-	      } else {
-		FD_CLR(process_itr->fd, &rfds_orig);
+	  }
+	  /* Remove the plugin */
+	  FD_CLR(proc->fd, &rfds_all);
+	  /* Block signal while modifying process_list */
+	  ret = sigprocmask (SIG_BLOCK, &sigchld_action.sa_mask, NULL);
+	  if(ret < 0){
+	    perror("sigprocmask");
+	    exitstatus = EXIT_FAILURE;
+	    goto end;
+	  }
+	  /* Delete this process entry from the list */
+	  if(process_list == proc){
+	    /* First one - simple */
+	    process_list = proc->next;
+	  } else {
+	    /* Second one or later */
+	    for(process *p = process_list; p != NULL; p = p->next){
+	      if(p->next == proc){
+		p->next = proc->next;
+		break;
 	      }
 	    }
 	  }
-	}
+	  /* We are done modifying process list, so unblock signal */
+	  ret = sigprocmask (SIG_UNBLOCK, &sigchld_action.sa_mask,
+			     NULL);
+	  if(ret < 0){
+	    perror("sigprocmask");
+	  }
+	  free(proc->buffer);
+	  free(proc);
+	  /* We deleted this process from the list, so we can't go
+	     proc->next.  Therefore, start over from the beginning of
+	     the process list */
+	  break;
+	}
+	/* This process exited nicely, so print its buffer */
+	for(size_t written = 0; written < proc->buffer_length;
+	    written += (size_t)ret){
+	  ret = TEMP_FAILURE_RETRY(write(STDOUT_FILENO,
+					 proc->buffer + written,
+					 proc->buffer_length
+					 - written));
+	  if(ret < 0){
+	    perror("write");
+	    exitstatus = EXIT_FAILURE;
+	    goto end;
+	  }
+	}
+	goto end;
+      }
+      /* This process has not completed.  Does it have any output? */
+      if(proc->eof or not FD_ISSET(proc->fd, &rfds)){
+	/* This process had nothing to say at this time */
+	continue;
+      }
+      /* Before reading, make the process' data buffer large enough */
+      if(proc->buffer_length + BUFFER_SIZE > proc->buffer_size){
+	proc->buffer = realloc(proc->buffer, proc->buffer_size
+			       + (size_t) BUFFER_SIZE);
+	if (proc->buffer == NULL){
+	  perror("malloc");
+	  exitstatus = EXIT_FAILURE;
+	  goto end;
+	}
+	proc->buffer_size += BUFFER_SIZE;
+      }
+      /* Read from the process */
+      ret = read(proc->fd, proc->buffer + proc->buffer_length,
+		 BUFFER_SIZE);
+      if(ret < 0){
+	/* Read error from this process; ignore the error */
+	continue;
+      }
+      if(ret == 0){
+	/* got EOF */
+	proc->eof = true;
+      } else {
+	proc->buffer_length += (size_t) ret;
       }
     }
   }
+  if(process_list == NULL){
+    fprintf(stderr, "All plugin processes failed, exiting\n");
+    exitstatus = EXIT_FAILURE;
+  }
   
  end:
-  for(process *process_itr = process_list; process_itr != NULL;
-      process_itr = process_itr->next){
-    close(process_itr->fd);
-    kill(process_itr->pid, SIGTERM);
-    free(process_itr->buffer);
-  }
-  
-  while(true){
-    int status;
-    ret = wait(&status);
-    if (ret == -1){
-      if(errno != ECHILD){
-	perror("wait");
-      }
-      break;
-    }
-  }  
-  return EXIT_SUCCESS;
+  /* Restore old signal handler */
+  sigaction(SIGCHLD, &old_sigchld_action, NULL);
+  
+  free(plus_argv);
+  
+  /* Free the plugin list */
+  for(plugin *next; plugin_list != NULL; plugin_list = next){
+    next = plugin_list->next;
+    free(plugin_list->argv);
+    free(plugin_list);
+  }
+  
+  if(dir != NULL){
+    closedir(dir);
+  }
+  
+  /* Free the process list and kill the processes */
+  for(process *next; process_list != NULL; process_list = next){
+    next = process_list->next;
+    close(process_list->fd);
+    kill(process_list->pid, SIGTERM);
+    free(process_list->buffer);
+    free(process_list);
+  }
+  
+  /* Wait for any remaining child processes to terminate */
+  do{
+    ret = wait(NULL);
+  } while(ret >= 0);
+  if(errno != ECHILD){
+    perror("wait");
+  }
+  
+  return exitstatus;
 }

=== removed file 'plugins.d/Makefile'
--- plugins.d/Makefile	2008-07-20 02:52:20 +0000
+++ plugins.d/Makefile	1970-01-01 00:00:00 +0000
@@ -1,18 +0,0 @@
-CFLAGS=-Wall -g --std=gnu99
-LDFLAGS=-lgnutls -lavahi-core -lgpgme
-
-PROGS=mandosclient passprompt
-
-objects=mandosclient.o passprompt.o
-
-all: $(PROGS)
-
-mandosclient: mandosclient.o
-	$(LINK.o) -lgnutls -lavahi-core -lgpgme $(COMMON) $^ $(LOADLIBES) $(LDLIBS) -o $@
-
-passprompt: passprompt.o
-	$(LINK.o) $(COMMON) $^ $(LOADLIBES) $(LDLIBS) -o $@
-
-.PHONY : clean
-clean :
-	-rm -f $(PROGS) $(objects) core

=== modified file 'plugins.d/mandosclient.c'
--- plugins.d/mandosclient.c	2008-07-20 22:58:26 +0000
+++ plugins.d/mandosclient.c	2008-08-04 16:05:01 +0000
@@ -1,30 +1,48 @@
-/***
-  This file is part of avahi.
- 
-  avahi is free software; you can redistribute it and/or modify it
-  under the terms of the GNU Lesser General Public License as
-  published by the Free Software Foundation; either version 2.1 of the
-  License, or (at your option) any later version.
- 
-  avahi is distributed in the hope that it will be useful, but WITHOUT
-  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
-  Public License for more details.
- 
-  You should have received a copy of the GNU Lesser General Public
-  License along with avahi; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
-  USA.
-***/
+/*  -*- coding: utf-8 -*- */
+/*
+ * Mandos client - get and decrypt data from a Mandos server
+ *
+ * This program is partly derived from an example program for an Avahi
+ * service browser, downloaded from
+ * <http://avahi.org/browser/examples/core-browse-services.c>.  This
+ * includes the following functions: "resolve_callback",
+ * "browse_callback", and parts of "main".
+ * 
+ * Everything else is
+ * Copyright © 2007-2008 Teddy Hogeborn & Björn Påhlsson
+ * 
+ * This program is free software: you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation, either version 3 of the
+ * License, or (at your option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, see
+ * <http://www.gnu.org/licenses/>.
+ * 
+ * Contact the authors at <mandos@fukt.bsnet.se>.
+ */
 
+/* Needed by GPGME, specifically gpgme_data_seek() */
 #define _LARGEFILE_SOURCE
 #define _FILE_OFFSET_BITS 64
 
+#define _GNU_SOURCE		/* TEMP_FAILURE_RETRY() */
+
 #include <stdio.h>
 #include <assert.h>
 #include <stdlib.h>
 #include <time.h>
 #include <net/if.h>		/* if_nametoindex */
+#include <sys/ioctl.h>          /* ioctl, ifreq, SIOCGIFFLAGS, IFF_UP,
+				   SIOCSIFFLAGS */
+#include <net/if.h>		/* ioctl, ifreq, SIOCGIFFLAGS, IFF_UP,
+				   SIOCSIFFLAGS */
 
 #include <avahi-core/core.h>
 #include <avahi-core/lookup.h>
@@ -33,11 +51,12 @@
 #include <avahi-common/malloc.h>
 #include <avahi-common/error.h>
 
-//mandos client part
-#include <sys/types.h>		/* socket(), setsockopt(), inet_pton() */
-#include <sys/socket.h>		/* socket(), setsockopt(), struct sockaddr_in6, struct in6_addr, inet_pton() */
-#include <gnutls/gnutls.h>	/* ALL GNUTLS STUFF */
-#include <gnutls/openpgp.h>	/* gnutls with openpgp stuff */
+/* Mandos client part */
+#include <sys/types.h>		/* socket(), inet_pton() */
+#include <sys/socket.h>		/* socket(), struct sockaddr_in6,
+				   struct in6_addr, inet_pton() */
+#include <gnutls/gnutls.h>	/* All GnuTLS stuff */
+#include <gnutls/openpgp.h>	/* GnuTLS with openpgp stuff */
 
 #include <unistd.h>		/* close() */
 #include <netinet/in.h>
@@ -45,47 +64,77 @@
 #include <string.h>		/* memset */
 #include <arpa/inet.h>		/* inet_pton() */
 #include <iso646.h>		/* not */
-
-// gpgme
+#include <net/if.h>		/* IF_NAMESIZE */
+#include <argp.h>		/* struct argp_option,
+				   struct argp_state, struct argp,
+				   argp_parse() */
+/* GPGME */
 #include <errno.h>		/* perror() */
 #include <gpgme.h>
 
-
-#ifndef CERT_ROOT
-#define CERT_ROOT "/conf/conf.d/cryptkeyreq/"
-#endif
-#define CERTFILE CERT_ROOT "openpgp-client.txt"
-#define KEYFILE CERT_ROOT "openpgp-client-key.txt"
 #define BUFFER_SIZE 256
-#define DH_BITS 1024
-
-bool debug;
-
+
+bool debug = false;
+static const char *keydir = "/conf/conf.d/mandos";
+const char *argp_program_version = "mandosclient 0.9";
+const char *argp_program_bug_address = "<mandos@fukt.bsnet.se>";
+static const char mandos_protocol_version[] = "1";
+
+/* Used for passing in values through the Avahi callback functions */
 typedef struct {
-  gnutls_session_t session;
+  AvahiSimplePoll *simple_poll;
+  AvahiServer *server;
   gnutls_certificate_credentials_t cred;
+  unsigned int dh_bits;
   gnutls_dh_params_t dh_params;
-} encrypted_session;
-
-
-ssize_t gpg_packet_decrypt (char *packet, size_t packet_size, char **new_packet, char *homedir){
+  const char *priority;
+} mandos_context;
+
+/* Make room in "buffer" for at least BUFFER_SIZE additional bytes.
+ * "buffer_capacity" is how much is currently allocated,
+ * "buffer_length" is how much is already used. */
+size_t adjustbuffer(char **buffer, size_t buffer_length,
+		  size_t buffer_capacity){
+  if (buffer_length + BUFFER_SIZE > buffer_capacity){
+    *buffer = realloc(*buffer, buffer_capacity + BUFFER_SIZE);
+    if (buffer == NULL){
+      return 0;
+    }
+    buffer_capacity += BUFFER_SIZE;
+  }
+  return buffer_capacity;
+}
+
+/* 
+ * Decrypt OpenPGP data using keyrings in HOMEDIR.
+ * Returns -1 on error
+ */
+static ssize_t pgp_packet_decrypt (const char *cryptotext,
+				   size_t crypto_size,
+				   char **plaintext,
+				   const char *homedir){
   gpgme_data_t dh_crypto, dh_plain;
   gpgme_ctx_t ctx;
   gpgme_error_t rc;
   ssize_t ret;
-  size_t new_packet_capacity = 0;
-  size_t new_packet_length = 0;
+  size_t plaintext_capacity = 0;
+  ssize_t plaintext_length = 0;
   gpgme_engine_info_t engine_info;
-
+  
   if (debug){
-    fprintf(stderr, "Attempting to decrypt password from gpg packet\n");
+    fprintf(stderr, "Trying to decrypt OpenPGP data\n");
   }
   
   /* Init GPGME */
   gpgme_check_version(NULL);
-  gpgme_engine_check_version(GPGME_PROTOCOL_OpenPGP);
+  rc = gpgme_engine_check_version(GPGME_PROTOCOL_OpenPGP);
+  if (rc != GPG_ERR_NO_ERROR){
+    fprintf(stderr, "bad gpgme_engine_check_version: %s: %s\n",
+	    gpgme_strsource(rc), gpgme_strerror(rc));
+    return -1;
+  }
   
-  /* Set GPGME home directory */
+  /* Set GPGME home directory for the OpenPGP engine only */
   rc = gpgme_get_engine_info (&engine_info);
   if (rc != GPG_ERR_NO_ERROR){
     fprintf(stderr, "bad gpgme_get_engine_info: %s: %s\n",
@@ -101,12 +150,13 @@
     engine_info = engine_info->next;
   }
   if(engine_info == NULL){
-    fprintf(stderr, "Could not set home dir to %s\n", homedir);
+    fprintf(stderr, "Could not set GPGME home dir to %s\n", homedir);
     return -1;
   }
   
-  /* Create new GPGME data buffer from packet buffer */
-  rc = gpgme_data_new_from_mem(&dh_crypto, packet, packet_size, 0);
+  /* Create new GPGME data buffer from memory cryptotext */
+  rc = gpgme_data_new_from_mem(&dh_crypto, cryptotext, crypto_size,
+			       0);
   if (rc != GPG_ERR_NO_ERROR){
     fprintf(stderr, "bad gpgme_data_new_from_mem: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
@@ -118,6 +168,7 @@
   if (rc != GPG_ERR_NO_ERROR){
     fprintf(stderr, "bad gpgme_data_new: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
+    gpgme_data_release(dh_crypto);
     return -1;
   }
   
@@ -126,29 +177,34 @@
   if (rc != GPG_ERR_NO_ERROR){
     fprintf(stderr, "bad gpgme_new: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
-    return -1;
+    plaintext_length = -1;
+    goto decrypt_end;
   }
   
-  /* Decrypt data from the FILE pointer to the plaintext data buffer */
+  /* Decrypt data from the cryptotext data buffer to the plaintext
+     data buffer */
   rc = gpgme_op_decrypt(ctx, dh_crypto, dh_plain);
   if (rc != GPG_ERR_NO_ERROR){
     fprintf(stderr, "bad gpgme_op_decrypt: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
-    return -1;
+    plaintext_length = -1;
+    goto decrypt_end;
   }
-
+  
   if(debug){
-    fprintf(stderr, "decryption of gpg packet succeeded\n");
+    fprintf(stderr, "Decryption of OpenPGP data succeeded\n");
   }
-
+  
   if (debug){
     gpgme_decrypt_result_t result;
     result = gpgme_op_decrypt_result(ctx);
     if (result == NULL){
       fprintf(stderr, "gpgme_op_decrypt_result failed\n");
     } else {
-      fprintf(stderr, "Unsupported algorithm: %s\n", result->unsupported_algorithm);
-      fprintf(stderr, "Wrong key usage: %d\n", result->wrong_key_usage);
+      fprintf(stderr, "Unsupported algorithm: %s\n",
+	      result->unsupported_algorithm);
+      fprintf(stderr, "Wrong key usage: %d\n",
+	      result->wrong_key_usage);
       if(result->file_name != NULL){
 	fprintf(stderr, "File name: %s\n", result->file_name);
       }
@@ -160,50 +216,63 @@
 		  gpgme_pubkey_algo_name(recipient->pubkey_algo));
 	  fprintf(stderr, "Key ID: %s\n", recipient->keyid);
 	  fprintf(stderr, "Secret key available: %s\n",
-		  recipient->status == GPG_ERR_NO_SECKEY ? "No" : "Yes");
+		  recipient->status == GPG_ERR_NO_SECKEY
+		  ? "No" : "Yes");
 	  recipient = recipient->next;
 	}
       }
     }
   }
   
-  /* Delete the GPGME FILE pointer cryptotext data buffer */
-  gpgme_data_release(dh_crypto);
-  
   /* Seek back to the beginning of the GPGME plaintext data buffer */
-  gpgme_data_seek(dh_plain, 0, SEEK_SET);
-
-  *new_packet = 0;
+  if (gpgme_data_seek(dh_plain, (off_t) 0, SEEK_SET) == -1){
+    perror("pgpme_data_seek");
+    plaintext_length = -1;
+    goto decrypt_end;
+  }
+  
+  *plaintext = NULL;
   while(true){
-    if (new_packet_length + BUFFER_SIZE > new_packet_capacity){
-      *new_packet = realloc(*new_packet, new_packet_capacity + BUFFER_SIZE);
-      if (*new_packet == NULL){
-	perror("realloc");
-	return -1;
-      }
-      new_packet_capacity += BUFFER_SIZE;
+    plaintext_capacity = adjustbuffer(plaintext,
+				      (size_t)plaintext_length,
+				      plaintext_capacity);
+    if (plaintext_capacity == 0){
+	perror("adjustbuffer");
+	plaintext_length = -1;
+	goto decrypt_end;
     }
     
-    ret = gpgme_data_read(dh_plain, *new_packet + new_packet_length, BUFFER_SIZE);
+    ret = gpgme_data_read(dh_plain, *plaintext + plaintext_length,
+			  BUFFER_SIZE);
     /* Print the data, if any */
     if (ret == 0){
-      /* If password is empty, then a incorrect error will be printed */
+      /* EOF */
       break;
     }
     if(ret < 0){
       perror("gpgme_data_read");
-      return -1;
+      plaintext_length = -1;
+      goto decrypt_end;
     }
-    new_packet_length += ret;
+    plaintext_length += ret;
   }
 
   if(debug){
-    fprintf(stderr, "decrypted password is: %s\n", *new_packet);
+    fprintf(stderr, "Decrypted password is: ");
+    for(ssize_t i = 0; i < plaintext_length; i++){
+      fprintf(stderr, "%02hhX ", (*plaintext)[i]);
+    }
+    fprintf(stderr, "\n");
   }
-
-   /* Delete the GPGME plaintext data buffer */
+  
+ decrypt_end:
+  
+  /* Delete the GPGME cryptotext data buffer */
+  gpgme_data_release(dh_crypto);
+  
+  /* Delete the GPGME plaintext data buffer */
   gpgme_data_release(dh_plain);
-  return new_packet_length;
+  return plaintext_length;
 }
 
 static const char * safer_gnutls_strerror (int value) {
@@ -213,119 +282,146 @@
   return ret;
 }
 
-void debuggnutls(int level, const char* string){
-  fprintf(stderr, "%s", string);
+/* GnuTLS log function callback */
+static void debuggnutls(__attribute__((unused)) int level,
+			const char* string){
+  fprintf(stderr, "GnuTLS: %s", string);
 }
 
-int initgnutls(encrypted_session *es){
-  const char *err;
+static int init_gnutls_global(mandos_context *mc,
+			      const char *pubkeyfile,
+			      const char *seckeyfile){
   int ret;
-
+  
   if(debug){
-    fprintf(stderr, "Initializing gnutls\n");
+    fprintf(stderr, "Initializing GnuTLS\n");
   }
 
-  
   if ((ret = gnutls_global_init ())
       != GNUTLS_E_SUCCESS) {
-    fprintf (stderr, "global_init: %s\n", safer_gnutls_strerror(ret));
+    fprintf (stderr, "GnuTLS global_init: %s\n",
+	     safer_gnutls_strerror(ret));
     return -1;
   }
-
+  
   if (debug){
+    /* "Use a log level over 10 to enable all debugging options."
+     * - GnuTLS manual
+     */
     gnutls_global_set_log_level(11);
     gnutls_global_set_log_function(debuggnutls);
   }
   
-
-  /* openpgp credentials */
-  if ((ret = gnutls_certificate_allocate_credentials (&es->cred))
+  /* OpenPGP credentials */
+  if ((ret = gnutls_certificate_allocate_credentials (&mc->cred))
       != GNUTLS_E_SUCCESS) {
-    fprintf (stderr, "memory error: %s\n", safer_gnutls_strerror(ret));
+    fprintf (stderr, "GnuTLS memory error: %s\n",
+	     safer_gnutls_strerror(ret));
     return -1;
   }
-
+  
   if(debug){
-    fprintf(stderr, "Attempting to use openpgp certificate %s"
-	    " and keyfile %s as gnutls credentials\n", CERTFILE, KEYFILE);
+    fprintf(stderr, "Attempting to use OpenPGP certificate %s"
+	    " and keyfile %s as GnuTLS credentials\n", pubkeyfile,
+	    seckeyfile);
   }
-
+  
   ret = gnutls_certificate_set_openpgp_key_file
-    (es->cred, CERTFILE, KEYFILE, GNUTLS_OPENPGP_FMT_BASE64);
-  if (ret != GNUTLS_E_SUCCESS) {
-    fprintf
-      (stderr, "Error[%d] while reading the OpenPGP key pair ('%s', '%s')\n",
-       ret, CERTFILE, KEYFILE);
-    fprintf(stdout, "The Error is: %s\n",
-	    safer_gnutls_strerror(ret));
-    return -1;
-  }
-
-  //Gnutls server initialization
-  if ((ret = gnutls_dh_params_init (&es->dh_params))
-      != GNUTLS_E_SUCCESS) {
-    fprintf (stderr, "Error in dh parameter initialization: %s\n",
-	     safer_gnutls_strerror(ret));
-    return -1;
-  }
-
-  if ((ret = gnutls_dh_params_generate2 (es->dh_params, DH_BITS))
-      != GNUTLS_E_SUCCESS) {
-    fprintf (stderr, "Error in prime generation: %s\n",
-	     safer_gnutls_strerror(ret));
-    return -1;
-  }
-
-  gnutls_certificate_set_dh_params (es->cred, es->dh_params);
-
-  // Gnutls session creation
-  if ((ret = gnutls_init (&es->session, GNUTLS_SERVER))
-      != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Error in gnutls session initialization: %s\n",
-	    safer_gnutls_strerror(ret));
-  }
-
-  if ((ret = gnutls_priority_set_direct (es->session, "NORMAL", &err))
-      != GNUTLS_E_SUCCESS) {
-    fprintf(stderr, "Syntax error at: %s\n", err);
-    fprintf(stderr, "Gnutls error: %s\n",
-	    safer_gnutls_strerror(ret));
-    return -1;
-  }
-
-  if ((ret = gnutls_credentials_set
-       (es->session, GNUTLS_CRD_CERTIFICATE, es->cred))
-      != GNUTLS_E_SUCCESS) {
-    fprintf(stderr, "Error setting a credentials set: %s\n",
-	    safer_gnutls_strerror(ret));
-    return -1;
-  }
-
+    (mc->cred, pubkeyfile, seckeyfile, GNUTLS_OPENPGP_FMT_BASE64);
+  if (ret != GNUTLS_E_SUCCESS) {
+    fprintf(stderr,
+	    "Error[%d] while reading the OpenPGP key pair ('%s',"
+	    " '%s')\n", ret, pubkeyfile, seckeyfile);
+    fprintf(stdout, "The GnuTLS error is: %s\n",
+	    safer_gnutls_strerror(ret));
+    return -1;
+  }
+  
+  /* GnuTLS server initialization */
+  ret = gnutls_dh_params_init(&mc->dh_params);
+  if (ret != GNUTLS_E_SUCCESS) {
+    fprintf (stderr, "Error in GnuTLS DH parameter initialization:"
+	     " %s\n", safer_gnutls_strerror(ret));
+    return -1;
+  }
+  ret = gnutls_dh_params_generate2(mc->dh_params, mc->dh_bits);
+  if (ret != GNUTLS_E_SUCCESS) {
+    fprintf (stderr, "Error in GnuTLS prime generation: %s\n",
+	     safer_gnutls_strerror(ret));
+    return -1;
+  }
+  
+  gnutls_certificate_set_dh_params(mc->cred, mc->dh_params);
+
+  return 0;
+}
+
+static int init_gnutls_session(mandos_context *mc,
+			       gnutls_session_t *session){
+  int ret;
+  /* GnuTLS session creation */
+  ret = gnutls_init(session, GNUTLS_SERVER);
+  if (ret != GNUTLS_E_SUCCESS){
+    fprintf(stderr, "Error in GnuTLS session initialization: %s\n",
+	    safer_gnutls_strerror(ret));
+  }
+  
+  {
+    const char *err;
+    ret = gnutls_priority_set_direct(*session, mc->priority, &err);
+    if (ret != GNUTLS_E_SUCCESS) {
+      fprintf(stderr, "Syntax error at: %s\n", err);
+      fprintf(stderr, "GnuTLS error: %s\n",
+	      safer_gnutls_strerror(ret));
+      return -1;
+    }
+  }
+  
+  ret = gnutls_credentials_set(*session, GNUTLS_CRD_CERTIFICATE,
+			       mc->cred);
+  if (ret != GNUTLS_E_SUCCESS) {
+    fprintf(stderr, "Error setting GnuTLS credentials: %s\n",
+	    safer_gnutls_strerror(ret));
+    return -1;
+  }
+  
   /* ignore client certificate if any. */
-  gnutls_certificate_server_set_request (es->session, GNUTLS_CERT_IGNORE);
+  gnutls_certificate_server_set_request (*session,
+					 GNUTLS_CERT_IGNORE);
   
-  gnutls_dh_set_prime_bits (es->session, DH_BITS);
+  gnutls_dh_set_prime_bits (*session, mc->dh_bits);
   
   return 0;
 }
 
-void empty_log(AvahiLogLevel level, const char *txt){}
+/* Avahi log function callback */
+static void empty_log(__attribute__((unused)) AvahiLogLevel level,
+		      __attribute__((unused)) const char *txt){}
 
-int start_mandos_communcation(char *ip, uint16_t port){
+/* Called when a Mandos server is found */
+static int start_mandos_communication(const char *ip, uint16_t port,
+				      AvahiIfIndex if_index,
+				      mandos_context *mc){
   int ret, tcp_sd;
-  struct sockaddr_in6 to;
-  struct in6_addr ip_addr;
-  encrypted_session es;
+  union { struct sockaddr in; struct sockaddr_in6 in6; } to;
   char *buffer = NULL;
   char *decrypted_buffer;
   size_t buffer_length = 0;
   size_t buffer_capacity = 0;
   ssize_t decrypted_buffer_size;
+  size_t written;
   int retval = 0;
-  const char interface[] = "eth0";
-
+  char interface[IF_NAMESIZE];
+  gnutls_session_t session;
+  
+  ret = init_gnutls_session (mc, &session);
+  if (ret != 0){
+    return -1;
+  }
+  
   if(debug){
-    fprintf(stderr, "Setting up a tcp connection to %s\n", ip);
+    fprintf(stderr, "Setting up a tcp connection to %s, port %d\n",
+	    ip, port);
   }
   
   tcp_sd = socket(PF_INET6, SOCK_STREAM, 0);
@@ -335,80 +431,108 @@
   }
 
   if(debug){
+    if(if_indextoname((unsigned int)if_index, interface) == NULL){
+      perror("if_indextoname");
+      return -1;
+    }
     fprintf(stderr, "Binding to interface %s\n", interface);
   }
-
-  ret = setsockopt(tcp_sd, SOL_SOCKET, SO_BINDTODEVICE, interface, 5);
-  if(tcp_sd < 0) {
-    perror("setsockopt bindtodevice");
-    return -1;
-  }
   
-  memset(&to,0,sizeof(to));
-  to.sin6_family = AF_INET6;
-  ret = inet_pton(AF_INET6, ip, &ip_addr);
+  memset(&to,0,sizeof(to));	/* Spurious warning */
+  to.in6.sin6_family = AF_INET6;
+  /* It would be nice to have a way to detect if we were passed an
+     IPv4 address here.   Now we assume an IPv6 address. */
+  ret = inet_pton(AF_INET6, ip, &to.in6.sin6_addr);
   if (ret < 0 ){
     perror("inet_pton");
     return -1;
-  }  
+  }
   if(ret == 0){
     fprintf(stderr, "Bad address: %s\n", ip);
     return -1;
   }
-  to.sin6_port = htons(port);
-  to.sin6_scope_id = if_nametoindex(interface);
-
+  to.in6.sin6_port = htons(port);	/* Spurious warning */
+  
+  to.in6.sin6_scope_id = (uint32_t)if_index;
+  
   if(debug){
-    fprintf(stderr, "Connection to: %s\n", ip);
+    fprintf(stderr, "Connection to: %s, port %d\n", ip, port);
+    char addrstr[INET6_ADDRSTRLEN] = "";
+    if(inet_ntop(to.in6.sin6_family, &(to.in6.sin6_addr), addrstr,
+		 sizeof(addrstr)) == NULL){
+      perror("inet_ntop");
+    } else {
+      if(strcmp(addrstr, ip) != 0){
+	fprintf(stderr, "Canonical address form: %s\n", addrstr);
+      }
+    }
   }
   
-  ret = connect(tcp_sd, (struct sockaddr *) &to, sizeof(to));
+  ret = connect(tcp_sd, &to.in, sizeof(to));
   if (ret < 0){
     perror("connect");
     return -1;
   }
-  
-  ret = initgnutls (&es);
-  if (ret != 0){
-    retval = -1;
-    return -1;
-  }
-    
-  
-  gnutls_transport_set_ptr (es.session, (gnutls_transport_ptr_t) tcp_sd);
 
+  const char *out = mandos_protocol_version;
+  written = 0;
+  while (true){
+    size_t out_size = strlen(out);
+    ret = TEMP_FAILURE_RETRY(write(tcp_sd, out + written,
+				   out_size - written));
+    if (ret == -1){
+      perror("write");
+      retval = -1;
+      goto mandos_end;
+    }
+    written += (size_t)ret;
+    if(written < out_size){
+      continue;
+    } else {
+      if (out == mandos_protocol_version){
+	written = 0;
+	out = "\r\n";
+      } else {
+	break;
+      }
+    }
+  }
+ 
   if(debug){
-    fprintf(stderr, "Establishing tls session with %s\n", ip);
+    fprintf(stderr, "Establishing TLS session with %s\n", ip);
   }
-
-  
-  ret = gnutls_handshake (es.session);
+  
+  gnutls_transport_set_ptr (session, (gnutls_transport_ptr_t) tcp_sd);
+  
+  ret = gnutls_handshake (session);
   
   if (ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "\n*** Handshake failed ***\n");
-    gnutls_perror (ret);
+    if(debug){
+      fprintf(stderr, "*** GnuTLS Handshake failed ***\n");
+      gnutls_perror (ret);
+    }
     retval = -1;
-    goto exit;
+    goto mandos_end;
   }
-
-  //Retrieve gpg packet that contains the wanted password
-
+  
+  /* Read OpenPGP packet that contains the wanted password */
+  
   if(debug){
-    fprintf(stderr, "Retrieving pgp encrypted password from %s\n", ip);
+    fprintf(stderr, "Retrieving pgp encrypted password from %s\n",
+	    ip);
   }
 
   while(true){
-    if (buffer_length + BUFFER_SIZE > buffer_capacity){
-      buffer = realloc(buffer, buffer_capacity + BUFFER_SIZE);
-      if (buffer == NULL){
-	perror("realloc");
-	goto exit;
-      }
-      buffer_capacity += BUFFER_SIZE;
+    buffer_capacity = adjustbuffer(&buffer, buffer_length,
+				   buffer_capacity);
+    if (buffer_capacity == 0){
+      perror("adjustbuffer");
+      retval = -1;
+      goto mandos_end;
     }
     
-    ret = gnutls_record_recv
-      (es.session, buffer+buffer_length, BUFFER_SIZE);
+    ret = gnutls_record_recv(session, buffer+buffer_length,
+			     BUFFER_SIZE);
     if (ret == 0){
       break;
     }
@@ -418,235 +542,462 @@
       case GNUTLS_E_AGAIN:
 	break;
       case GNUTLS_E_REHANDSHAKE:
-	ret = gnutls_handshake (es.session);
+	ret = gnutls_handshake (session);
 	if (ret < 0){
-	  fprintf(stderr, "\n*** Handshake failed ***\n");
+	  fprintf(stderr, "*** GnuTLS Re-handshake failed ***\n");
 	  gnutls_perror (ret);
 	  retval = -1;
-	  goto exit;
+	  goto mandos_end;
 	}
 	break;
       default:
-	fprintf(stderr, "Unknown error while reading data from encrypted session with mandos server\n");
+	fprintf(stderr, "Unknown error while reading data from"
+		" encrypted session with Mandos server\n");
 	retval = -1;
-	gnutls_bye (es.session, GNUTLS_SHUT_RDWR);
-	goto exit;
+	gnutls_bye (session, GNUTLS_SHUT_RDWR);
+	goto mandos_end;
       }
     } else {
-      buffer_length += ret;
+      buffer_length += (size_t) ret;
     }
   }
   
+  if(debug){
+    fprintf(stderr, "Closing TLS session\n");
+  }
+  
+  gnutls_bye (session, GNUTLS_SHUT_RDWR);
+  
   if (buffer_length > 0){
-    if ((decrypted_buffer_size = gpg_packet_decrypt(buffer, buffer_length, &decrypted_buffer, CERT_ROOT)) >= 0){
-      fwrite (decrypted_buffer, 1, decrypted_buffer_size, stdout);
+    decrypted_buffer_size = pgp_packet_decrypt(buffer,
+					       buffer_length,
+					       &decrypted_buffer,
+					       keydir);
+    if (decrypted_buffer_size >= 0){
+      written = 0;
+      while(written < (size_t) decrypted_buffer_size){
+	ret = (int)fwrite (decrypted_buffer + written, 1,
+			   (size_t)decrypted_buffer_size - written,
+			   stdout);
+	if(ret == 0 and ferror(stdout)){
+	  if(debug){
+	    fprintf(stderr, "Error writing encrypted data: %s\n",
+		    strerror(errno));
+	  }
+	  retval = -1;
+	  break;
+	}
+	written += (size_t)ret;
+      }
       free(decrypted_buffer);
     } else {
       retval = -1;
     }
   }
-
-  //shutdown procedure
-
-  if(debug){
-    fprintf(stderr, "Closing tls session\n");
-  }
-
+  
+  /* Shutdown procedure */
+  
+ mandos_end:
   free(buffer);
-  gnutls_bye (es.session, GNUTLS_SHUT_RDWR);
- exit:
   close(tcp_sd);
-  gnutls_deinit (es.session);
-  gnutls_certificate_free_credentials (es.cred);
+  gnutls_deinit (session);
+  gnutls_certificate_free_credentials (mc->cred);
   gnutls_global_deinit ();
   return retval;
 }
 
-static AvahiSimplePoll *simple_poll = NULL;
-static AvahiServer *server = NULL;
-
-static void resolve_callback(
-    AvahiSServiceResolver *r,
-    AVAHI_GCC_UNUSED AvahiIfIndex interface,
-    AVAHI_GCC_UNUSED AvahiProtocol protocol,
-    AvahiResolverEvent event,
-    const char *name,
-    const char *type,
-    const char *domain,
-    const char *host_name,
-    const AvahiAddress *address,
-    uint16_t port,
-    AvahiStringList *txt,
-    AvahiLookupResultFlags flags,
-    AVAHI_GCC_UNUSED void* userdata) {
-    
-    assert(r);
-
-    /* Called whenever a service has been resolved successfully or timed out */
-
-    switch (event) {
-        case AVAHI_RESOLVER_FAILURE:
-            fprintf(stderr, "(Resolver) Failed to resolve service '%s' of type '%s' in domain '%s': %s\n", name, type, domain, avahi_strerror(avahi_server_errno(server)));
-            break;
-
-        case AVAHI_RESOLVER_FOUND: {
-	  char ip[AVAHI_ADDRESS_STR_MAX];
-            avahi_address_snprint(ip, sizeof(ip), address);
-	    if(debug){
-	      fprintf(stderr, "Mandos server found at %s on port %d\n", ip, port);
-	    }
-	    int ret = start_mandos_communcation(ip, port);
-	    if (ret == 0){
-	      exit(EXIT_SUCCESS);
-	    } else {
-	      exit(EXIT_FAILURE);
-	    }
-        }
-    }
-    avahi_s_service_resolver_free(r);
-}
-
-static void browse_callback(
-    AvahiSServiceBrowser *b,
-    AvahiIfIndex interface,
-    AvahiProtocol protocol,
-    AvahiBrowserEvent event,
-    const char *name,
-    const char *type,
-    const char *domain,
-    AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
-    void* userdata) {
-    
-    AvahiServer *s = userdata;
-    assert(b);
-
-    /* Called whenever a new services becomes available on the LAN or is removed from the LAN */
-
-    switch (event) {
-
-        case AVAHI_BROWSER_FAILURE:
-            
-            fprintf(stderr, "(Browser) %s\n", avahi_strerror(avahi_server_errno(server)));
-            avahi_simple_poll_quit(simple_poll);
-            return;
-
-        case AVAHI_BROWSER_NEW:
-            /* We ignore the returned resolver object. In the callback
-               function we free it. If the server is terminated before
-               the callback function is called the server will free
-               the resolver for us. */
-            
-            if (!(avahi_s_service_resolver_new(s, interface, protocol, name, type, domain, AVAHI_PROTO_INET6, 0, resolve_callback, s)))
-                fprintf(stderr, "Failed to resolve service '%s': %s\n", name, avahi_strerror(avahi_server_errno(s)));
-            
-            break;
-
-        case AVAHI_BROWSER_REMOVE:
-            break;
-
-        case AVAHI_BROWSER_ALL_FOR_NOW:
-        case AVAHI_BROWSER_CACHE_EXHAUSTED:
-            break;
-    }
-}
-
-int main(AVAHI_GCC_UNUSED int argc, AVAHI_GCC_UNUSED char*argv[]) {
-    AvahiServerConfig config;
+static void resolve_callback(AvahiSServiceResolver *r,
+			     AvahiIfIndex interface,
+			     AVAHI_GCC_UNUSED AvahiProtocol protocol,
+			     AvahiResolverEvent event,
+			     const char *name,
+			     const char *type,
+			     const char *domain,
+			     const char *host_name,
+			     const AvahiAddress *address,
+			     uint16_t port,
+			     AVAHI_GCC_UNUSED AvahiStringList *txt,
+			     AVAHI_GCC_UNUSED AvahiLookupResultFlags
+			     flags,
+			     void* userdata) {
+  mandos_context *mc = userdata;
+  assert(r);			/* Spurious warning */
+  
+  /* Called whenever a service has been resolved successfully or
+     timed out */
+  
+  switch (event) {
+  default:
+  case AVAHI_RESOLVER_FAILURE:
+    fprintf(stderr, "(Avahi Resolver) Failed to resolve service '%s'"
+	    " of type '%s' in domain '%s': %s\n", name, type, domain,
+	    avahi_strerror(avahi_server_errno(mc->server)));
+    break;
+    
+  case AVAHI_RESOLVER_FOUND:
+    {
+      char ip[AVAHI_ADDRESS_STR_MAX];
+      avahi_address_snprint(ip, sizeof(ip), address);
+      if(debug){
+	fprintf(stderr, "Mandos server \"%s\" found on %s (%s, %d) on"
+		" port %d\n", name, host_name, ip, interface, port);
+      }
+      int ret = start_mandos_communication(ip, port, interface, mc);
+      if (ret == 0){
+	exit(EXIT_SUCCESS);
+      }
+    }
+  }
+  avahi_s_service_resolver_free(r);
+}
+
+static void browse_callback( AvahiSServiceBrowser *b,
+			     AvahiIfIndex interface,
+			     AvahiProtocol protocol,
+			     AvahiBrowserEvent event,
+			     const char *name,
+			     const char *type,
+			     const char *domain,
+			     AVAHI_GCC_UNUSED AvahiLookupResultFlags
+			     flags,
+			     void* userdata) {
+  mandos_context *mc = userdata;
+  assert(b);			/* Spurious warning */
+  
+  /* Called whenever a new services becomes available on the LAN or
+     is removed from the LAN */
+  
+  switch (event) {
+  default:
+  case AVAHI_BROWSER_FAILURE:
+    
+    fprintf(stderr, "(Avahi browser) %s\n",
+	    avahi_strerror(avahi_server_errno(mc->server)));
+    avahi_simple_poll_quit(mc->simple_poll);
+    return;
+    
+  case AVAHI_BROWSER_NEW:
+    /* We ignore the returned Avahi resolver object. In the callback
+       function we free it. If the Avahi server is terminated before
+       the callback function is called the Avahi server will free the
+       resolver for us. */
+    
+    if (!(avahi_s_service_resolver_new(mc->server, interface,
+				       protocol, name, type, domain,
+				       AVAHI_PROTO_INET6, 0,
+				       resolve_callback, mc)))
+      fprintf(stderr, "Avahi: Failed to resolve service '%s': %s\n",
+	      name, avahi_strerror(avahi_server_errno(mc->server)));
+    break;
+    
+  case AVAHI_BROWSER_REMOVE:
+    break;
+    
+  case AVAHI_BROWSER_ALL_FOR_NOW:
+  case AVAHI_BROWSER_CACHE_EXHAUSTED:
+    if(debug){
+      fprintf(stderr, "No Mandos server found, still searching...\n");
+    }
+    break;
+  }
+}
+
+/* Combines file name and path and returns the malloced new
+   string. some sane checks could/should be added */
+static const char *combinepath(const char *first, const char *second){
+  size_t f_len = strlen(first);
+  size_t s_len = strlen(second);
+  char *tmp = malloc(f_len + s_len + 2);
+  if (tmp == NULL){
+    return NULL;
+  }
+  if(f_len > 0){
+    memcpy(tmp, first, f_len);	/* Spurious warning */
+  }
+  tmp[f_len] = '/';
+  if(s_len > 0){
+    memcpy(tmp + f_len + 1, second, s_len); /* Spurious warning */
+  }
+  tmp[f_len + 1 + s_len] = '\0';
+  return tmp;
+}
+
+
+int main(int argc, char *argv[]){
     AvahiSServiceBrowser *sb = NULL;
-    const char db[] = "--debug";
     int error;
-    int ret = 1;
-    int returncode = EXIT_SUCCESS;
-    char *basename = rindex(argv[0], '/');
-    if(basename == NULL){
-      basename = argv[0];
-    } else {
-      basename++;
-    }
+    int ret;
+    int exitcode = EXIT_SUCCESS;
+    const char *interface = "eth0";
+    struct ifreq network;
+    int sd;
+    uid_t uid;
+    gid_t gid;
+    char *connect_to = NULL;
+    AvahiIfIndex if_index = AVAHI_IF_UNSPEC;
+    const char *pubkeyfile = "pubkey.txt";
+    const char *seckeyfile = "seckey.txt";
+    mandos_context mc = { .simple_poll = NULL, .server = NULL,
+			  .dh_bits = 1024, .priority = "SECURE256"};
     
-    char *program_name = malloc(strlen(basename) + sizeof(db));
+    {
+      struct argp_option options[] = {
+	{ .name = "debug", .key = 128,
+	  .doc = "Debug mode", .group = 3 },
+	{ .name = "connect", .key = 'c',
+	  .arg = "IP",
+	  .doc = "Connect directly to a sepcified mandos server",
+	  .group = 1 },
+	{ .name = "interface", .key = 'i',
+	  .arg = "INTERFACE",
+	  .doc = "Interface that Avahi will conntect through",
+	  .group = 1 },
+	{ .name = "keydir", .key = 'd',
+	  .arg = "KEYDIR",
+	  .doc = "Directory where the openpgp keyring is",
+	  .group = 1 },
+	{ .name = "seckey", .key = 's',
+	  .arg = "SECKEY",
+	  .doc = "Secret openpgp key for gnutls authentication",
+	  .group = 1 },
+	{ .name = "pubkey", .key = 'p',
+	  .arg = "PUBKEY",
+	  .doc = "Public openpgp key for gnutls authentication",
+	  .group = 2 },
+	{ .name = "dh-bits", .key = 129,
+	  .arg = "BITS",
+	  .doc = "dh-bits to use in gnutls communication",
+	  .group = 2 },
+	{ .name = "priority", .key = 130,
+	  .arg = "PRIORITY",
+	  .doc = "GNUTLS priority", .group = 1 },
+	{ .name = NULL }
+      };
 
-    if (program_name == NULL){
-      perror("argv[0]");
-      return EXIT_FAILURE;
-    }
-    
-    program_name[0] = '\0';
-    
-    for (int i = 1; i < argc; i++){
-      if (not strncmp(argv[i], db, 5)){
-	  strcat(strcat(strcat(program_name, db ), "="), basename);
-	  if(not strcmp(argv[i], db) or not strcmp(argv[i], program_name)){
-	    debug = true;
+      
+      error_t parse_opt (int key, char *arg,
+			 struct argp_state *state) {
+	/* Get the INPUT argument from `argp_parse', which we know is
+	   a pointer to our plugin list pointer. */
+	switch (key) {
+	case 128:
+	  debug = true;
+	  break;
+	case 'c':
+	  connect_to = arg;
+	  break;
+	case 'i':
+	  interface = arg;
+	  break;
+	case 'd':
+	  keydir = arg;
+	  break;
+	case 's':
+	  seckeyfile = arg;
+	  break;
+	case 'p':
+	  pubkeyfile = arg;
+	  break;
+	case 129:
+	  errno = 0;
+	  mc.dh_bits = (unsigned int) strtol(arg, NULL, 10);
+	  if (errno){
+	    perror("strtol");
+	    exit(EXIT_FAILURE);
 	  }
-	}
-    }
-    free(program_name);
-
+	  break;
+	case 130:
+	  mc.priority = arg;
+	  break;
+	case ARGP_KEY_ARG:
+	  argp_usage (state);
+	  break;
+	  case ARGP_KEY_END:
+	    break;
+	default:
+	  return ARGP_ERR_UNKNOWN;
+	}
+	return 0;
+      }
+
+      struct argp argp = { .options = options, .parser = parse_opt,
+			   .args_doc = "",
+			   .doc = "Mandos client -- Get and decrypt"
+			   " passwords from mandos server" };
+      argp_parse (&argp, argc, argv, 0, 0, NULL);
+    }
+      
+    pubkeyfile = combinepath(keydir, pubkeyfile);
+    if (pubkeyfile == NULL){
+      perror("combinepath");
+      exitcode = EXIT_FAILURE;
+      goto end;
+    }
+    
+    seckeyfile = combinepath(keydir, seckeyfile);
+    if (seckeyfile == NULL){
+      perror("combinepath");
+      goto end;
+    }
+
+    ret = init_gnutls_global(&mc, pubkeyfile, seckeyfile);
+    if (ret == -1){
+      fprintf(stderr, "init_gnutls_global\n");
+      goto end;
+    }
+
+    uid = getuid();
+    gid = getgid();
+
+    ret = setuid(uid);
+    if (ret == -1){
+      perror("setuid");
+    }
+    
+    setgid(gid);
+    if (ret == -1){
+      perror("setgid");
+    }
+    
+    if_index = (AvahiIfIndex) if_nametoindex(interface);
+    if(if_index == 0){
+      fprintf(stderr, "No such interface: \"%s\"\n", interface);
+      exit(EXIT_FAILURE);
+    }
+    
+    if(connect_to != NULL){
+      /* Connect directly, do not use Zeroconf */
+      /* (Mainly meant for debugging) */
+      char *address = strrchr(connect_to, ':');
+      if(address == NULL){
+        fprintf(stderr, "No colon in address\n");
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+      errno = 0;
+      uint16_t port = (uint16_t) strtol(address+1, NULL, 10);
+      if(errno){
+	perror("Bad port number");
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+      *address = '\0';
+      address = connect_to;
+      ret = start_mandos_communication(address, port, if_index, &mc);
+      if(ret < 0){
+	exitcode = EXIT_FAILURE;
+      } else {
+	exitcode = EXIT_SUCCESS;
+      }
+      goto end;
+    }
+    
+    /* If the interface is down, bring it up */
+    {
+      sd = socket(PF_INET6, SOCK_DGRAM, IPPROTO_IP);
+      if(sd < 0) {
+	perror("socket");
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+      strcpy(network.ifr_name, interface); /* Spurious warning */
+      ret = ioctl(sd, SIOCGIFFLAGS, &network);
+      if(ret == -1){
+	perror("ioctl SIOCGIFFLAGS");
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+      if((network.ifr_flags & IFF_UP) == 0){
+	network.ifr_flags |= IFF_UP;
+	ret = ioctl(sd, SIOCSIFFLAGS, &network);
+	if(ret == -1){
+	  perror("ioctl SIOCSIFFLAGS");
+	  exitcode = EXIT_FAILURE;
+	  goto end;
+	}
+      }
+      close(sd);
+    }
+    
     if (not debug){
       avahi_set_log_function(empty_log);
     }
     
-    /* Initialize the psuedo-RNG */
-    srand(time(NULL));
-
-    /* Allocate main loop object */
-    if (!(simple_poll = avahi_simple_poll_new())) {
-        fprintf(stderr, "Failed to create simple poll object.\n");
-	
-        goto exit;
-    }
-
-    /* Do not publish any local records */
-    avahi_server_config_init(&config);
-    config.publish_hinfo = 0;
-    config.publish_addresses = 0;
-    config.publish_workstation = 0;
-    config.publish_domain = 0;
-
-    /* Allocate a new server */
-    server = avahi_server_new(avahi_simple_poll_get(simple_poll), &config, NULL, NULL, &error);
-
-    /* Free the configuration data */
-    avahi_server_config_free(&config);
-
-    /* Check if creating the server object succeeded */
-    if (!server) {
-        fprintf(stderr, "Failed to create server: %s\n", avahi_strerror(error));
-	returncode = EXIT_FAILURE;
-        goto exit;
-    }
-    
-    /* Create the service browser */
-    if (!(sb = avahi_s_service_browser_new(server, if_nametoindex("eth0"), AVAHI_PROTO_INET6, "_mandos._tcp", NULL, 0, browse_callback, server))) {
-        fprintf(stderr, "Failed to create service browser: %s\n", avahi_strerror(avahi_server_errno(server)));
-	returncode = EXIT_FAILURE;
-        goto exit;
+    /* Initialize the pseudo-RNG for Avahi */
+    srand((unsigned int) time(NULL));
+    
+    /* Allocate main Avahi loop object */
+    mc.simple_poll = avahi_simple_poll_new();
+    if (mc.simple_poll == NULL) {
+        fprintf(stderr, "Avahi: Failed to create simple poll"
+		" object.\n");
+	exitcode = EXIT_FAILURE;
+        goto end;
+    }
+
+    {
+      AvahiServerConfig config;
+      /* Do not publish any local Zeroconf records */
+      avahi_server_config_init(&config);
+      config.publish_hinfo = 0;
+      config.publish_addresses = 0;
+      config.publish_workstation = 0;
+      config.publish_domain = 0;
+
+      /* Allocate a new server */
+      mc.server = avahi_server_new(avahi_simple_poll_get
+				   (mc.simple_poll), &config, NULL,
+				   NULL, &error);
+    
+      /* Free the Avahi configuration data */
+      avahi_server_config_free(&config);
+    }
+    
+    /* Check if creating the Avahi server object succeeded */
+    if (mc.server == NULL) {
+        fprintf(stderr, "Failed to create Avahi server: %s\n",
+		avahi_strerror(error));
+	exitcode = EXIT_FAILURE;
+        goto end;
+    }
+    
+    /* Create the Avahi service browser */
+    sb = avahi_s_service_browser_new(mc.server, if_index,
+				     AVAHI_PROTO_INET6,
+				     "_mandos._tcp", NULL, 0,
+				     browse_callback, &mc);
+    if (sb == NULL) {
+        fprintf(stderr, "Failed to create service browser: %s\n",
+		avahi_strerror(avahi_server_errno(mc.server)));
+	exitcode = EXIT_FAILURE;
+        goto end;
     }
     
     /* Run the main loop */
 
     if (debug){
-      fprintf(stderr, "Starting avahi loop search\n");
+      fprintf(stderr, "Starting Avahi loop search\n");
     }
     
-    avahi_simple_poll_loop(simple_poll);
+    avahi_simple_poll_loop(mc.simple_poll);
     
-exit:
+ end:
 
     if (debug){
       fprintf(stderr, "%s exiting\n", argv[0]);
     }
     
     /* Cleanup things */
-    if (sb)
+    if (sb != NULL)
         avahi_s_service_browser_free(sb);
     
-    if (server)
-        avahi_server_free(server);
-
-    if (simple_poll)
-        avahi_simple_poll_free(simple_poll);
-
-    return ret;
+    if (mc.server != NULL)
+        avahi_server_free(mc.server);
+
+    if (mc.simple_poll != NULL)
+        avahi_simple_poll_free(mc.simple_poll);
+    free(pubkeyfile);
+    free(seckeyfile);
+    
+    return exitcode;
 }

=== modified file 'plugins.d/passprompt.c'
--- plugins.d/passprompt.c	2008-07-20 02:52:20 +0000
+++ plugins.d/passprompt.c	2008-08-04 16:23:54 +0000
@@ -1,5 +1,29 @@
+/*  -*- coding: utf-8 -*- */
+/*
+ * Passprompt - Read a password from the terminal and print it
+ *
+ * Copyright © 2007-2008 Teddy Hogeborn & Björn Påhlsson
+ * 
+ * This program is free software: you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation, either version 3 of the
+ * License, or (at your option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, see
+ * <http://www.gnu.org/licenses/>.
+ * 
+ * Contact the authors at <https://www.fukt.bsnet.se/~belorn/> and
+ * <https://www.fukt.bsnet.se/~teddy/>.
+ */
+
 #define _GNU_SOURCE		/* getline() */
-#define _FORTIFY_SOURCE 2
+
 #include <termios.h> 		/* struct termios, tcsetattr(),
 				   TCSAFLUSH, tcgetattr(), ECHO */
 #include <unistd.h>		/* struct termios, tcsetattr(),
@@ -11,29 +35,81 @@
 				   SIGQUIT, SIGHUP, SIGTERM */
 #include <stddef.h>		/* NULL, size_t */
 #include <sys/types.h>		/* ssize_t */
-#include <stdlib.h>		/* EXIT_SUCCESS, EXIT_FAILURE */
+#include <stdlib.h>		/* EXIT_SUCCESS, EXIT_FAILURE,
+				   getopt_long */
 #include <stdio.h>		/* fprintf(), stderr, getline(),
 				   stdin, feof(), perror(), fputc(),
-				   stdout */
+				   stdout, getopt_long */
 #include <errno.h>		/* errno, EINVAL */
 #include <iso646.h>		/* or, not */
 #include <stdbool.h>		/* bool, false, true */
+#include <string.h> 		/* strlen, rindex, strncmp, strcmp */
+#include <argp.h>		/* struct argp_option,
+				   struct argp_state, struct argp,
+				   argp_parse() */
 
 volatile bool quit_now = false;
+bool debug = false;
+const char *argp_program_version = "passprompt 0.9";
+const char *argp_program_bug_address = "<mandos@fukt.bsnet.se>";
 
-void termination_handler(int signum){
+static void termination_handler(__attribute__((unused))int signum){
   quit_now = true;
 }
 
 int main(int argc, char **argv){
-  ssize_t ret = -1;
+  ssize_t ret;
   size_t n;
   struct termios t_new, t_old;
   char *buffer = NULL;
+  char *prefix = NULL;
   int status = EXIT_SUCCESS;
   struct sigaction old_action,
     new_action = { .sa_handler = termination_handler,
 		   .sa_flags = 0 };
+  {
+    struct argp_option options[] = {
+      { .name = "prefix", .key = 'p',
+	.arg = "PREFIX", .flags = 0,
+	.doc = "Prefix used before the passprompt", .group = 2 },
+      { .name = "debug", .key = 128,
+	.doc = "Debug mode", .group = 3 },
+      { .name = NULL }
+    };
+  
+    error_t parse_opt (int key, char *arg, struct argp_state *state) {
+      /* Get the INPUT argument from `argp_parse', which we know is a
+	 pointer to our plugin list pointer. */
+      switch (key) {
+      case 'p':
+	prefix = arg;
+	break;
+      case 128:
+	debug = true;
+	break;
+      case ARGP_KEY_ARG:
+	argp_usage (state);
+	break;
+      case ARGP_KEY_END:
+	break;
+      default:
+	return ARGP_ERR_UNKNOWN;
+      }
+      return 0;
+    }
+  
+    struct argp argp = { .options = options, .parser = parse_opt,
+			 .args_doc = "",
+			 .doc = "Mandos Passprompt -- Provides a passprompt" };
+    argp_parse (&argp, argc, argv, 0, 0, NULL);
+  }
+    
+  if (debug){
+    fprintf(stderr, "Starting %s\n", argv[0]);
+  }
+  if (debug){
+    fprintf(stderr, "Storing current terminal attributes\n");
+  }
   
   if (tcgetattr(STDIN_FILENO, &t_old) != 0){
     return EXIT_FAILURE;
@@ -41,21 +117,22 @@
   
   sigemptyset(&new_action.sa_mask);
   sigaddset(&new_action.sa_mask, SIGINT);
-  sigaddset(&new_action.sa_mask, SIGQUIT);
   sigaddset(&new_action.sa_mask, SIGHUP);
   sigaddset(&new_action.sa_mask, SIGTERM);
   sigaction(SIGINT, NULL, &old_action);
   if (old_action.sa_handler != SIG_IGN)
     sigaction(SIGINT, &new_action, NULL);
-  sigaction(SIGQUIT, NULL, &old_action);
-  if (old_action.sa_handler != SIG_IGN)
-    sigaction(SIGQUIT, &new_action, NULL);
   sigaction(SIGHUP, NULL, &old_action);
   if (old_action.sa_handler != SIG_IGN)
     sigaction(SIGHUP, &new_action, NULL);
   sigaction(SIGTERM, NULL, &old_action);
   if (old_action.sa_handler != SIG_IGN)
     sigaction(SIGTERM, &new_action, NULL);
+
+  
+  if (debug){
+    fprintf(stderr, "Removing echo flag from terminal attributes\n");
+  }
   
   t_new = t_old;
   t_new.c_lflag &= ~ECHO;
@@ -63,20 +140,27 @@
     perror("tcsetattr-echo");
     return EXIT_FAILURE;
   }
-  
+
+  if (debug){
+    fprintf(stderr, "Waiting for input from stdin \n");
+  }
   while(true){
     if (quit_now){
       status = EXIT_FAILURE;
       break;
     }
-    fprintf(stderr, "Password: ");
+
+    if(prefix){
+      fprintf(stderr, "%s Password: ", prefix);
+    } else {
+      fprintf(stderr, "Password: ");
+    }      
     ret = getline(&buffer, &n, stdin);
     if (ret > 0){
       fprintf(stdout, "%s", buffer);
       status = EXIT_SUCCESS;
       break;
     }
-    // ret == 0 makes no other sence than to retry to read from stdin
     if (ret < 0){
       if (errno != EINTR and not feof(stdin)){
 	perror("getline");
@@ -84,12 +168,21 @@
 	break;
       }
     }
+    /* if(ret == 0), then the only sensible thing to do is to retry to
+       read from stdin */
     fputc('\n', stderr);
   }
-
+  
+  if (debug){
+    fprintf(stderr, "Restoring terminal attributes\n");
+  }
   if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_old) != 0){
     perror("tcsetattr+echo");
   }
   
+  if (debug){
+    fprintf(stderr, "%s is exiting\n", argv[0]);
+  }
+  
   return status;
 }

=== added file 'server.conf'
--- server.conf	1970-01-01 00:00:00 +0000
+++ server.conf	2008-07-29 03:35:39 +0000
@@ -0,0 +1,7 @@
+[server]
+#interface = 
+#address =
+#port = 
+#debug = 
+#priority = SECURE256:!CTYPE-X.509:+CTYPE-OPENPGP
+#servicename = Mandos

=== modified file 'server.py'
--- server.py	2008-07-20 06:33:48 +0000
+++ server.py	2008-08-03 21:45:14 +0000
@@ -1,5 +1,33 @@
 #!/usr/bin/python
 # -*- mode: python; coding: utf-8 -*-
+# 
+# Mandos server - give out binary blobs to connecting clients.
+# 
+# This program is partly derived from an example program for an Avahi
+# service publisher, downloaded from
+# <http://avahi.org/wiki/PythonPublishExample>.  This includes the
+# following functions: "AvahiService.add", "AvahiService.remove",
+# "server_state_changed", "entry_group_state_changed", and some lines
+# in "main".
+# 
+# Everything else is
+# Copyright © 2007-2008 Teddy Hogeborn & Björn Påhlsson
+# 
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+#     This program is distributed in the hope that it will be useful,
+#     but WITHOUT ANY WARRANTY; without even the implied warranty of
+#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+#     GNU General Public License for more details.
+# 
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# 
+# Contact the authors at <mandos@fukt.bsnet.se>.
+# 
 
 from __future__ import division
 
@@ -24,6 +52,8 @@
 import subprocess
 import atexit
 import stat
+import logging
+import logging.handlers
 
 import dbus
 import gobject
@@ -31,8 +61,17 @@
 from dbus.mainloop.glib import DBusGMainLoop
 import ctypes
 
-import logging
-import logging.handlers
+# Brief description of the operation of this program:
+# 
+# This server announces itself as a Zeroconf service.  Connecting
+# clients use the TLS protocol, with the unusual quirk that this
+# server program acts as a TLS "client" while a connecting client acts
+# as a TLS "server".  The client (acting as a TLS "server") must
+# supply an OpenPGP certificate, and the fingerprint of this
+# certificate is used by this server to look up (in a list read from a
+# file at start time) which binary blob to give the client.  No other
+# authentication or authorization is done by this server.
+
 
 logger = logging.Logger('mandos')
 syslogger = logging.handlers.SysLogHandler\
@@ -42,20 +81,90 @@
 logger.addHandler(syslogger)
 del syslogger
 
-# This variable is used to optionally bind to a specified interface.
-# It is a global variable to fit in with the other variables from the
-# Avahi server example code.
-serviceInterface = avahi.IF_UNSPEC
-# From the Avahi server example code:
-serviceName = "Mandos"
-serviceType = "_mandos._tcp" # http://www.dns-sd.org/ServiceTypes.html
-servicePort = None                      # Not known at startup
-serviceTXT = []                         # TXT record for the service
-domain = ""                  # Domain to publish on, default to .local
-host = ""          # Host to publish records for, default to localhost
-group = None #our entry group
-rename_count = 12       # Counter so we only rename after collisions a
-                        # sensible number of times
+
+class AvahiError(Exception):
+    def __init__(self, value):
+        self.value = value
+    def __str__(self):
+        return repr(self.value)
+
+class AvahiServiceError(AvahiError):
+    pass
+
+class AvahiGroupError(AvahiError):
+    pass
+
+
+class AvahiService(object):
+    """
+    interface: integer; avahi.IF_UNSPEC or an interface index.
+               Used to optionally bind to the specified interface.
+    name = string; Example: "Mandos"
+    type = string; Example: "_mandos._tcp".
+                   See <http://www.dns-sd.org/ServiceTypes.html>
+    port = integer; what port to announce
+    TXT = list of strings; TXT record for the service
+    domain = string; Domain to publish on, default to .local if empty.
+    host = string; Host to publish records for, default to localhost
+                   if empty.
+    max_renames = integer; maximum number of renames
+    rename_count = integer; counter so we only rename after collisions
+                   a sensible number of times
+    """
+    def __init__(self, interface = avahi.IF_UNSPEC, name = None,
+                 type = None, port = None, TXT = None, domain = "",
+                 host = "", max_renames = 12):
+        """An Avahi (Zeroconf) service. """
+        self.interface = interface
+        self.name = name
+        self.type = type
+        self.port = port
+        if TXT is None:
+            self.TXT = []
+        else:
+            self.TXT = TXT
+        self.domain = domain
+        self.host = host
+        self.rename_count = 0
+    def rename(self):
+        """Derived from the Avahi example code"""
+        if self.rename_count >= self.max_renames:
+            logger.critical(u"No suitable service name found after %i"
+                            u" retries, exiting.", rename_count)
+            raise AvahiServiceError("Too many renames")
+        name = server.GetAlternativeServiceName(name)
+        logger.notice(u"Changing name to %r ...", name)
+        self.remove()
+        self.add()
+        self.rename_count += 1
+    def remove(self):
+        """Derived from the Avahi example code"""
+        if group is not None:
+            group.Reset()
+    def add(self):
+        """Derived from the Avahi example code"""
+        global group
+        if group is None:
+            group = dbus.Interface\
+                    (bus.get_object(avahi.DBUS_NAME,
+                                    server.EntryGroupNew()),
+                     avahi.DBUS_INTERFACE_ENTRY_GROUP)
+            group.connect_to_signal('StateChanged',
+                                    entry_group_state_changed)
+        logger.debug(u"Adding service '%s' of type '%s' ...",
+                     service.name, service.type)
+        group.AddService(
+                self.interface,         # interface
+                avahi.PROTO_INET6,      # protocol
+                dbus.UInt32(0),         # flags
+                self.name, self.type,
+                self.domain, self.host,
+                dbus.UInt16(self.port),
+                avahi.string_array_to_txt_array(self.TXT))
+        group.Commit()
+
+# From the Avahi example code:
+group = None                            # our entry group
 # End of Avahi example code
 
 
@@ -67,26 +176,26 @@
                  uniquely identify the client
     secret:    bytestring; sent verbatim (over TLS) to client
     fqdn:      string (FQDN); available for use by the checker command
-    created:   datetime.datetime()
-    last_seen: datetime.datetime() or None if not yet seen
-    timeout:   datetime.timedelta(); How long from last_seen until
-                                     this client is invalid
+    created:   datetime.datetime(); object creation, not client host
+    last_checked_ok: datetime.datetime() or None if not yet checked OK
+    timeout:   datetime.timedelta(); How long from last_checked_ok
+                                     until this client is invalid
     interval:  datetime.timedelta(); How often to start a new checker
     stop_hook: If set, called by stop() as stop_hook(self)
     checker:   subprocess.Popen(); a running checker process used
                                    to see if the client lives.
-                                   Is None if no process is running.
+                                   'None' if no process is running.
     checker_initiator_tag: a gobject event source tag, or None
     stop_initiator_tag:    - '' -
     checker_callback_tag:  - '' -
     checker_command: string; External command which is run to check if
-                     client lives.  %()s expansions are done at
+                     client lives.  %() expansions are done at
                      runtime with vars(self) as dict, so that for
                      instance %(name)s can be used in the command.
     Private attibutes:
     _timeout: Real variable for 'timeout'
     _interval: Real variable for 'interval'
-    _timeout_milliseconds: Used by gobject.timeout_add()
+    _timeout_milliseconds: Used when calling gobject.timeout_add()
     _interval_milliseconds: - '' -
     """
     def _set_timeout(self, timeout):
@@ -112,14 +221,18 @@
     interval = property(lambda self: self._interval,
                         _set_interval)
     del _set_interval
-    def __init__(self, name=None, options=None, stop_hook=None,
-                 fingerprint=None, secret=None, secfile=None,
-                 fqdn=None, timeout=None, interval=-1, checker=None):
+    def __init__(self, name=None, stop_hook=None, fingerprint=None,
+                 secret=None, secfile=None, fqdn=None, timeout=None,
+                 interval=-1, checker=None):
+        """Note: the 'checker' argument sets the 'checker_command'
+        attribute and not the 'checker' attribute.."""
         self.name = name
+        logger.debug(u"Creating client %r", self.name)
         # Uppercase and remove spaces from fingerprint
         # for later comparison purposes with return value of
         # the fingerprint() function
         self.fingerprint = fingerprint.upper().replace(u" ", u"")
+        logger.debug(u"  Fingerprint: %s", self.fingerprint)
         if secret:
             self.secret = secret.decode(u"base64")
         elif secfile:
@@ -127,19 +240,13 @@
             self.secret = sf.read()
             sf.close()
         else:
-            raise RuntimeError(u"No secret or secfile for client %s"
-                               % self.name)
-        self.fqdn = fqdn                # string
+            raise TypeError(u"No secret or secfile for client %s"
+                            % self.name)
+        self.fqdn = fqdn
         self.created = datetime.datetime.now()
-        self.last_seen = None
-        if timeout is None:
-            self.timeout = options.timeout
-        else:
-            self.timeout = string_to_delta(timeout)
-        if interval == -1:
-            self.interval = options.interval
-        else:
-            self.interval = string_to_delta(interval)
+        self.last_checked_ok = None
+        self.timeout = string_to_delta(timeout)
+        self.interval = string_to_delta(interval)
         self.stop_hook = stop_hook
         self.checker = None
         self.checker_initiator_tag = None
@@ -147,7 +254,7 @@
         self.checker_callback_tag = None
         self.check_command = checker
     def start(self):
-        """Start this clients checker and timeout hooks"""
+        """Start this client's checker and timeout hooks"""
         # Schedule a new checker to be started an 'interval' from now,
         # and every interval from then on.
         self.checker_initiator_tag = gobject.timeout_add\
@@ -161,14 +268,18 @@
                                    self.stop)
     def stop(self):
         """Stop this client.
-        The possibility that this client might be restarted is left
-        open, but not currently used."""
-        logger.debug(u"Stopping client %s", self.name)
-        self.secret = None
-        if self.stop_initiator_tag:
+        The possibility that a client might be restarted is left open,
+        but not currently used."""
+        # If this client doesn't have a secret, it is already stopped.
+        if self.secret:
+            logger.debug(u"Stopping client %s", self.name)
+            self.secret = None
+        else:
+            return False
+        if getattr(self, "stop_initiator_tag", False):
             gobject.source_remove(self.stop_initiator_tag)
             self.stop_initiator_tag = None
-        if self.checker_initiator_tag:
+        if getattr(self, "checker_initiator_tag", False):
             gobject.source_remove(self.checker_initiator_tag)
             self.checker_initiator_tag = None
         self.stop_checker()
@@ -177,24 +288,18 @@
         # Do not run this again if called by a gobject.timeout_add
         return False
     def __del__(self):
-        # Some code duplication here and in stop()
-        if hasattr(self, "stop_initiator_tag") \
-               and self.stop_initiator_tag:
-            gobject.source_remove(self.stop_initiator_tag)
-            self.stop_initiator_tag = None
-        if hasattr(self, "checker_initiator_tag") \
-               and self.checker_initiator_tag:
-            gobject.source_remove(self.checker_initiator_tag)
-            self.checker_initiator_tag = None
-        self.stop_checker()
+        self.stop_hook = None
+        self.stop()
     def checker_callback(self, pid, condition):
         """The checker has completed, so take appropriate actions."""
         now = datetime.datetime.now()
+        self.checker_callback_tag = None
+        self.checker = None
         if os.WIFEXITED(condition) \
                and (os.WEXITSTATUS(condition) == 0):
             logger.debug(u"Checker for %(name)s succeeded",
                          vars(self))
-            self.last_seen = now
+            self.last_checked_ok = now
             gobject.source_remove(self.stop_initiator_tag)
             self.stop_initiator_tag = gobject.timeout_add\
                                       (self._timeout_milliseconds,
@@ -205,32 +310,39 @@
         else:
             logger.debug(u"Checker for %(name)s failed",
                          vars(self))
-            self.checker = None
-        self.checker_callback_tag = None
     def start_checker(self):
         """Start a new checker subprocess if one is not running.
         If a checker already exists, leave it running and do
         nothing."""
+        # The reason for not killing a running checker is that if we
+        # did that, then if a checker (for some reason) started
+        # running slowly and taking more than 'interval' time, the
+        # client would inevitably timeout, since no checker would get
+        # a chance to run to completion.  If we instead leave running
+        # checkers alone, the checker would have to take more time
+        # than 'timeout' for the client to be declared invalid, which
+        # is as it should be.
         if self.checker is None:
             try:
+                # In case check_command has exactly one % operator
                 command = self.check_command % self.fqdn
             except TypeError:
+                # Escape attributes for the shell
                 escaped_attrs = dict((key, re.escape(str(val)))
                                      for key, val in
                                      vars(self).iteritems())
                 try:
                     command = self.check_command % escaped_attrs
                 except TypeError, error:
-                    logger.critical(u'Could not format string "%s":'
-                                    u' %s', self.check_command, error)
+                    logger.error(u'Could not format string "%s":'
+                                 u' %s', self.check_command, error)
                     return True # Try again later
             try:
                 logger.debug(u"Starting checker %r for %s",
                              command, self.name)
-                self.checker = subprocess.\
-                               Popen(command,
-                                     close_fds=True, shell=True,
-                                     cwd="/")
+                self.checker = subprocess.Popen(command,
+                                                close_fds=True,
+                                                shell=True, cwd="/")
                 self.checker_callback_tag = gobject.child_watch_add\
                                             (self.checker.pid,
                                              self.checker_callback)
@@ -241,26 +353,32 @@
         return True
     def stop_checker(self):
         """Force the checker process, if any, to stop."""
-        if not hasattr(self, "checker") or self.checker is None:
+        if self.checker_callback_tag:
+            gobject.source_remove(self.checker_callback_tag)
+            self.checker_callback_tag = None
+        if getattr(self, "checker", None) is None:
             return
-        gobject.source_remove(self.checker_callback_tag)
-        self.checker_callback_tag = None
-        os.kill(self.checker.pid, signal.SIGTERM)
-        if self.checker.poll() is None:
-            os.kill(self.checker.pid, signal.SIGKILL)
+        logger.debug("Stopping checker for %(name)s", vars(self))
+        try:
+            os.kill(self.checker.pid, signal.SIGTERM)
+            #os.sleep(0.5)
+            #if self.checker.poll() is None:
+            #    os.kill(self.checker.pid, signal.SIGKILL)
+        except OSError, error:
+            if error.errno != errno.ESRCH: # No such process
+                raise
         self.checker = None
-    def still_valid(self, now=None):
+    def still_valid(self):
         """Has the timeout not yet passed for this client?"""
-        if now is None:
-            now = datetime.datetime.now()
-        if self.last_seen is None:
+        now = datetime.datetime.now()
+        if self.last_checked_ok is None:
             return now < (self.created + self.timeout)
         else:
-            return now < (self.last_seen + self.timeout)
+            return now < (self.last_checked_ok + self.timeout)
 
 
 def peer_certificate(session):
-    "Return an OpenPGP data packet string for the peer's certificate"
+    "Return the peer's OpenPGP certificate as a bytestring"
     # If not an OpenPGP certificate...
     if gnutls.library.functions.gnutls_certificate_type_get\
             (session._c_object) \
@@ -277,7 +395,7 @@
 
 
 def fingerprint(openpgp):
-    "Convert an OpenPGP data string to a hexdigit fingerprint string"
+    "Convert an OpenPGP bytestring to a hexdigit fingerprint string"
     # New empty GnuTLS certificate
     crt = gnutls.library.types.gnutls_openpgp_crt_t()
     gnutls.library.functions.gnutls_openpgp_crt_init\
@@ -315,15 +433,30 @@
     def handle(self):
         logger.debug(u"TCP connection from: %s",
                      unicode(self.client_address))
-        session = gnutls.connection.ClientSession(self.request,
-                                                  gnutls.connection.\
-                                                  X509Credentials())
+        session = gnutls.connection.ClientSession\
+                  (self.request, gnutls.connection.X509Credentials())
+        
+        line = self.request.makefile().readline()
+        logger.debug(u"Protocol version: %r", line)
+        try:
+            if int(line.strip().split()[0]) > 1:
+                raise RuntimeError
+        except (ValueError, IndexError, RuntimeError), error:
+            logger.error(u"Unknown protocol version: %s", error)
+            return
+        
+        # Note: gnutls.connection.X509Credentials is really a generic
+        # GnuTLS certificate credentials object so long as no X.509
+        # keys are added to it.  Therefore, we can use it here despite
+        # using OpenPGP certificates.
         
         #priority = ':'.join(("NONE", "+VERS-TLS1.1", "+AES-256-CBC",
         #                "+SHA1", "+COMP-NULL", "+CTYPE-OPENPGP",
         #                "+DHE-DSS"))
-        priority = "SECURE256"
-        
+        priority = "NORMAL"             # Fallback default, since this
+                                        # MUST be set.
+        if self.server.settings["priority"]:
+            priority = self.server.settings["priority"]
         gnutls.library.functions.gnutls_priority_set_direct\
             (session._c_object, priority, None);
         
@@ -342,20 +475,19 @@
             return
         logger.debug(u"Fingerprint: %s", fpr)
         client = None
-        for c in clients:
+        for c in self.server.clients:
             if c.fingerprint == fpr:
                 client = c
                 break
+        if not client:
+            logger.debug(u"Client not found for fingerprint: %s", fpr)
+            session.bye()
+            return
         # Have to check if client.still_valid(), since it is possible
         # that the client timed out while establishing the GnuTLS
         # session.
-        if (not client) or (not client.still_valid()):
-            if client:
-                logger.debug(u"Client %(name)s is invalid",
-                             vars(client))
-            else:
-                logger.debug(u"Client not found for fingerprint: %s",
-                             fpr)
+        if not client.still_valid():
+            logger.debug(u"Client %(name)s is invalid", vars(client))
             session.bye()
             return
         sent_size = 0
@@ -371,14 +503,14 @@
 class IPv6_TCPServer(SocketServer.ForkingTCPServer, object):
     """IPv6 TCP server.  Accepts 'None' as address and/or port.
     Attributes:
-        options:        Command line options
+        settings:       Server settings
         clients:        Set() of Client objects
     """
     address_family = socket.AF_INET6
     def __init__(self, *args, **kwargs):
-        if "options" in kwargs:
-            self.options = kwargs["options"]
-            del kwargs["options"]
+        if "settings" in kwargs:
+            self.settings = kwargs["settings"]
+            del kwargs["settings"]
         if "clients" in kwargs:
             self.clients = kwargs["clients"]
             del kwargs["clients"]
@@ -387,19 +519,18 @@
         """This overrides the normal server_bind() function
         to bind to an interface if one was specified, and also NOT to
         bind to an address or port if they were not specified."""
-        if self.options.interface:
-            if not hasattr(socket, "SO_BINDTODEVICE"):
-                # From /usr/include/asm-i486/socket.h
-                socket.SO_BINDTODEVICE = 25
+        if self.settings["interface"]:
+            # 25 is from /usr/include/asm-i486/socket.h
+            SO_BINDTODEVICE = getattr(socket, "SO_BINDTODEVICE", 25)
             try:
                 self.socket.setsockopt(socket.SOL_SOCKET,
-                                       socket.SO_BINDTODEVICE,
-                                       self.options.interface)
+                                       SO_BINDTODEVICE,
+                                       self.settings["interface"])
             except socket.error, error:
                 if error[0] == errno.EPERM:
                     logger.warning(u"No permission to"
                                    u" bind to interface %s",
-                                   self.options.interface)
+                                   self.settings["interface"])
                 else:
                     raise error
         # Only bind(2) the socket if we really need to.
@@ -448,93 +579,52 @@
     return delta
 
 
-def add_service():
-    """From the Avahi server example code"""
-    global group, serviceName, serviceType, servicePort, serviceTXT, \
-           domain, host
-    if group is None:
-        group = dbus.Interface(
-                bus.get_object( avahi.DBUS_NAME,
-                                server.EntryGroupNew()),
-                avahi.DBUS_INTERFACE_ENTRY_GROUP)
-        group.connect_to_signal('StateChanged',
-                                entry_group_state_changed)
-    logger.debug(u"Adding service '%s' of type '%s' ...",
-                 serviceName, serviceType)
-    
-    group.AddService(
-            serviceInterface,           # interface
-            avahi.PROTO_INET6,          # protocol
-            dbus.UInt32(0),             # flags
-            serviceName, serviceType,
-            domain, host,
-            dbus.UInt16(servicePort),
-            avahi.string_array_to_txt_array(serviceTXT))
-    group.Commit()
-
-
-def remove_service():
-    """From the Avahi server example code"""
-    global group
-    
-    if not group is None:
-        group.Reset()
-
-
 def server_state_changed(state):
-    """From the Avahi server example code"""
+    """Derived from the Avahi example code"""
     if state == avahi.SERVER_COLLISION:
         logger.warning(u"Server name collision")
-        remove_service()
+        service.remove()
     elif state == avahi.SERVER_RUNNING:
-        add_service()
+        service.add()
 
 
 def entry_group_state_changed(state, error):
-    """From the Avahi server example code"""
-    global serviceName, server, rename_count
-    
+    """Derived from the Avahi example code"""
     logger.debug(u"state change: %i", state)
     
     if state == avahi.ENTRY_GROUP_ESTABLISHED:
         logger.debug(u"Service established.")
     elif state == avahi.ENTRY_GROUP_COLLISION:
-        
-        rename_count = rename_count - 1
-        if rename_count > 0:
-            name = server.GetAlternativeServiceName(name)
-            logger.warning(u"Service name collision, "
-                           u"changing name to '%s' ...", name)
-            remove_service()
-            add_service()
-            
-        else:
-            logger.error(u"No suitable service name found after %i"
-                         u" retries, exiting.", n_rename)
-            killme(1)
+        logger.warning(u"Service name collision.")
+        service.rename()
     elif state == avahi.ENTRY_GROUP_FAILURE:
-        logger.error(u"Error in group state changed %s",
-                     unicode(error))
-        killme(1)
-
+        logger.critical(u"Error in group state changed %s",
+                        unicode(error))
+        raise AvahiGroupError("State changed: %s", str(error))
 
 def if_nametoindex(interface):
-    """Call the C function if_nametoindex()"""
+    """Call the C function if_nametoindex(), or equivalent"""
+    global if_nametoindex
     try:
-        libc = ctypes.cdll.LoadLibrary("libc.so.6")
-        return libc.if_nametoindex(interface)
+        if "ctypes.util" not in sys.modules:
+            import ctypes.util
+        if_nametoindex = ctypes.cdll.LoadLibrary\
+            (ctypes.util.find_library("c")).if_nametoindex
     except (OSError, AttributeError):
         if "struct" not in sys.modules:
             import struct
         if "fcntl" not in sys.modules:
             import fcntl
-        SIOCGIFINDEX = 0x8933      # From /usr/include/linux/sockios.h
-        s = socket.socket()
-        ifreq = fcntl.ioctl(s, SIOCGIFINDEX,
-                            struct.pack("16s16x", interface))
-        s.close()
-        interface_index = struct.unpack("I", ifreq[16:20])[0]
-        return interface_index
+        def if_nametoindex(interface):
+            "Get an interface index the hard way, i.e. using fcntl()"
+            SIOCGIFINDEX = 0x8933  # From /usr/include/linux/sockios.h
+            s = socket.socket()
+            ifreq = fcntl.ioctl(s, SIOCGIFINDEX,
+                                struct.pack("16s16x", interface))
+            s.close()
+            interface_index = struct.unpack("I", ifreq[16:20])[0]
+            return interface_index
+    return if_nametoindex(interface)
 
 
 def daemon(nochdir, noclose):
@@ -547,7 +637,7 @@
         os.chdir("/")
     if not noclose:
         # Close all standard open file descriptors
-        null = os.open("/dev/null", os.O_NOCTTY | os.O_RDWR)
+        null = os.open(os.path.devnull, os.O_NOCTTY | os.O_RDWR)
         if not stat.S_ISCHR(os.fstat(null).st_mode):
             raise OSError(errno.ENODEV,
                           "/dev/null not a character device")
@@ -558,34 +648,30 @@
             os.close(null)
 
 
-def killme(status = 0):
-    logger.debug("Stopping server with exit status %d", status)
-    exitstatus = status
-    if main_loop_started:
-        main_loop.quit()
-    else:
-        sys.exit(status)
-
-
-if __name__ == '__main__':
-    exitstatus = 0
+def main():
+    global main_loop_started
     main_loop_started = False
+    
     parser = OptionParser()
     parser.add_option("-i", "--interface", type="string",
-                      default=None, metavar="IF",
-                      help="Bind to interface IF")
-    parser.add_option("-p", "--port", type="int", default=None,
+                      metavar="IF", help="Bind to interface IF")
+    parser.add_option("-a", "--address", type="string",
+                      help="Address to listen for requests on")
+    parser.add_option("-p", "--port", type="int",
                       help="Port number to receive requests on")
-    parser.add_option("--timeout", type="string", # Parsed later
-                      default="1h",
-                      help="Amount of downtime allowed for clients")
-    parser.add_option("--interval", type="string", # Parsed later
-                      default="5m",
-                      help="How often to check that a client is up")
     parser.add_option("--check", action="store_true", default=False,
                       help="Run self-test")
     parser.add_option("--debug", action="store_true", default=False,
-                      help="Debug mode")
+                      help="Debug mode; run in foreground and log to"
+                      " terminal")
+    parser.add_option("--priority", type="string", help="GnuTLS"
+                      " priority string (see GnuTLS documentation)")
+    parser.add_option("--servicename", type="string", metavar="NAME",
+                      help="Zeroconf service name")
+    parser.add_option("--configdir", type="string",
+                      default="/etc/mandos", metavar="DIR",
+                      help="Directory to search for configuration"
+                      " files")
     (options, args) = parser.parse_args()
     
     if options.check:
@@ -593,23 +679,57 @@
         doctest.testmod()
         sys.exit()
     
-    # Parse the time arguments
-    try:
-        options.timeout = string_to_delta(options.timeout)
-    except ValueError:
-        parser.error("option --timeout: Unparseable time")
-    try:
-        options.interval = string_to_delta(options.interval)
-    except ValueError:
-        parser.error("option --interval: Unparseable time")
-    
-    # Parse config file
-    defaults = { "checker": "fping -q -- %%(fqdn)s" }
-    client_config = ConfigParser.SafeConfigParser(defaults)
-    #client_config.readfp(open("secrets.conf"), "secrets.conf")
-    client_config.read("mandos-clients.conf")
-    
-    # From the Avahi server example code
+    # Default values for config file for server-global settings
+    server_defaults = { "interface": "",
+                        "address": "",
+                        "port": "",
+                        "debug": "False",
+                        "priority":
+                        "SECURE256:!CTYPE-X.509:+CTYPE-OPENPGP",
+                        "servicename": "Mandos",
+                        }
+    
+    # Parse config file for server-global settings
+    server_config = ConfigParser.SafeConfigParser(server_defaults)
+    del server_defaults
+    server_config.read(os.path.join(options.configdir, "server.conf"))
+    server_section = "server"
+    # Convert the SafeConfigParser object to a dict
+    server_settings = dict(server_config.items(server_section))
+    # Use getboolean on the boolean config option
+    server_settings["debug"] = server_config.getboolean\
+                               (server_section, "debug")
+    del server_config
+    
+    # Override the settings from the config file with command line
+    # options, if set.
+    for option in ("interface", "address", "port", "debug",
+                   "priority", "servicename", "configdir"):
+        value = getattr(options, option)
+        if value is not None:
+            server_settings[option] = value
+    del options
+    # Now we have our good server settings in "server_settings"
+    
+    # Parse config file with clients
+    client_defaults = { "timeout": "1h",
+                        "interval": "5m",
+                        "checker": "fping -q -- %%(fqdn)s",
+                        }
+    client_config = ConfigParser.SafeConfigParser(client_defaults)
+    client_config.read(os.path.join(server_settings["configdir"],
+                                    "clients.conf"))
+    
+    global service
+    service = AvahiService(name = server_settings["servicename"],
+                           type = "_mandos._tcp", );
+    if server_settings["interface"]:
+        service.interface = if_nametoindex(server_settings["interface"])
+    
+    global main_loop
+    global bus
+    global server
+    # From the Avahi example code
     DBusGMainLoop(set_as_default=True )
     main_loop = gobject.MainLoop()
     bus = dbus.SystemBus()
@@ -618,7 +738,7 @@
             avahi.DBUS_INTERFACE_SERVER )
     # End of Avahi example code
     
-    debug = options.debug
+    debug = server_settings["debug"]
     
     if debug:
         console = logging.StreamHandler()
@@ -633,9 +753,9 @@
         clients.remove(client)
         if not clients:
             logger.debug(u"No clients left, exiting")
-            killme()
+            sys.exit()
     
-    clients.update(Set(Client(name=section, options=options,
+    clients.update(Set(Client(name=section,
                               stop_hook = remove_from_clients,
                               **(dict(client_config\
                                       .items(section))))
@@ -647,13 +767,14 @@
     def cleanup():
         "Cleanup function; run on exit"
         global group
-        # From the Avahi server example code
+        # From the Avahi example code
         if not group is None:
             group.Free()
             group = None
         # End of Avahi example code
         
-        for client in clients:
+        while clients:
+            client = clients.pop()
             client.stop_hook = None
             client.stop()
     
@@ -661,41 +782,48 @@
     
     if not debug:
         signal.signal(signal.SIGINT, signal.SIG_IGN)
-    signal.signal(signal.SIGHUP, lambda signum, frame: killme())
-    signal.signal(signal.SIGTERM, lambda signum, frame: killme())
+    signal.signal(signal.SIGHUP, lambda signum, frame: sys.exit())
+    signal.signal(signal.SIGTERM, lambda signum, frame: sys.exit())
     
     for client in clients:
         client.start()
     
-    tcp_server = IPv6_TCPServer((None, options.port),
+    tcp_server = IPv6_TCPServer((server_settings["address"],
+                                 server_settings["port"]),
                                 tcp_handler,
-                                options=options,
+                                settings=server_settings,
                                 clients=clients)
-    # Find out what random port we got
-    servicePort = tcp_server.socket.getsockname()[1]
-    logger.debug(u"Now listening on port %d", servicePort)
-    
-    if options.interface is not None:
-        serviceInterface = if_nametoindex(options.interface)
-    
-    # From the Avahi server example code
-    server.connect_to_signal("StateChanged", server_state_changed)
-    try:
-        server_state_changed(server.GetState())
-    except dbus.exceptions.DBusException, error:
-        logger.critical(u"DBusException: %s", error)
-        killme(1)
-    # End of Avahi example code
-    
-    gobject.io_add_watch(tcp_server.fileno(), gobject.IO_IN,
-                         lambda *args, **kwargs:
-                         tcp_server.handle_request(*args[2:],
-                                                   **kwargs) or True)
-    try:
+    # Find out what port we got
+    service.port = tcp_server.socket.getsockname()[1]
+    logger.debug(u"Now listening on address %r, port %d, flowinfo %d,"
+                 u" scope_id %d" % tcp_server.socket.getsockname())
+    
+    #service.interface = tcp_server.socket.getsockname()[3]
+    
+    try:
+        # From the Avahi example code
+        server.connect_to_signal("StateChanged", server_state_changed)
+        try:
+            server_state_changed(server.GetState())
+        except dbus.exceptions.DBusException, error:
+            logger.critical(u"DBusException: %s", error)
+            sys.exit(1)
+        # End of Avahi example code
+        
+        gobject.io_add_watch(tcp_server.fileno(), gobject.IO_IN,
+                             lambda *args, **kwargs:
+                             tcp_server.handle_request\
+                             (*args[2:], **kwargs) or True)
+        
+        logger.debug("Starting main loop")
         main_loop_started = True
         main_loop.run()
+    except AvahiError, error:
+        logger.critical(u"AvahiError: %s" + unicode(error))
+        sys.exit(1)
     except KeyboardInterrupt:
         if debug:
             print
-    
-    sys.exit(exitstatus)
+
+if __name__ == '__main__':
+    main()

